#
# Copyright (C) 2011 Nick Lanham <nick@afternight.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from deluge.ui.client import client
from deluge.ui.console.utils import curses_util as util
from deluge.ui.console.widgets.popup import MessagePopup, SelectablePopup

from . import ACTION

try:
    import curses
except ImportError:
    pass

key_to_action = {
    curses.KEY_HOME: ACTION.QUEUE_TOP,
    curses.KEY_UP: ACTION.QUEUE_UP,
    curses.KEY_DOWN: ACTION.QUEUE_DOWN,
    curses.KEY_END: ACTION.QUEUE_BOTTOM,
}
QUEUE_MODE_HELP_STR = """
Change queue position of selected torrents

{!info!}'+'{!normal!} - {|indent_pos:|}Move up
{!info!}'-'{!normal!} - {|indent_pos:|}Move down

{!info!}'Home'{!normal!} - {|indent_pos:|}Move to top
{!info!}'End'{!normal!} - {|indent_pos:|}Move to bottom

"""


class QueueMode:
    def __init__(self, torrentslist, torrent_ids):
        self.torrentslist = torrentslist
        self.torrentview = torrentslist.torrentview
        self.torrent_ids = torrent_ids

    def set_statusbar_args(self, statusbar_args):
        statusbar_args[
            'bottombar'
        ] = '{!black,white!}Queue mode: change queue position of selected torrents.'
        statusbar_args['bottombar_help'] = ' Press [h] for help'

    def update_cursor(self):
        pass

    def update_colors(self, tidx, colors):
        pass

    def handle_read(self, c):
        if c in [util.KEY_ESC, util.KEY_BELL]:  # If Escape key or CTRL-g, we abort
            self.torrentslist.set_minor_mode(None)
        elif c == ord('h'):
            popup = MessagePopup(
                self.torrentslist,
                'Help',
                QUEUE_MODE_HELP_STR,
                width_req=0.65,
                border_off_west=1,
            )
            self.torrentslist.push_popup(popup, clear=True)
        elif c in [
            curses.KEY_UP,
            curses.KEY_DOWN,
            curses.KEY_HOME,
            curses.KEY_END,
            curses.KEY_NPAGE,
            curses.KEY_PPAGE,
        ]:
            action = key_to_action[c]
            self.do_queue(action)

    def move_selection(self, cb_arg, qact):
        if self.torrentslist.config['torrentview']['move_selection'] is False:
            return
        queue_length = 0
        selected_num = 0
        for tid in self.torrentview.curstate:
            tq = self.torrentview.curstate[tid]['queue']
            if tq != -1:
                queue_length += 1
                if tq in self.torrentview.marked:
                    selected_num += 1
        if qact == ACTION.QUEUE_TOP:
            if self.torrentview.marked:
                self.torrentview.cursel = 1 + sorted(self.torrentview.marked).index(
                    self.torrentview.cursel
                )
            else:
                self.torrentview.cursel = 1
            self.torrentview.marked = list(range(1, selected_num + 1))
        elif qact == ACTION.QUEUE_UP:
            self.torrentview.cursel = max(1, self.torrentview.cursel - 1)
            self.torrentview.marked = [marked - 1 for marked in self.torrentview.marked]
            self.torrentview.marked = [
                marked for marked in self.torrentview.marked if marked > 0
            ]
        elif qact == ACTION.QUEUE_DOWN:
            self.torrentview.cursel = min(queue_length, self.torrentview.cursel + 1)
            self.torrentview.marked = [marked + 1 for marked in self.torrentview.marked]
            self.torrentview.marked = [
                marked for marked in self.torrentview.marked if marked <= queue_length
            ]
        elif qact == ACTION.QUEUE_BOTTOM:
            if self.torrentview.marked:
                self.torrentview.cursel = (
                    queue_length
                    - selected_num
                    + 1
                    + sorted(self.torrentview.marked).index(self.torrentview.cursel)
                )
            else:
                self.torrentview.cursel = queue_length
            self.torrentview.marked = list(
                range(queue_length - selected_num + 1, queue_length + 1)
            )

    def do_queue(self, qact, *args, **kwargs):
        if qact == ACTION.QUEUE_TOP:
            client.core.queue_top(self.torrent_ids).addCallback(
                self.move_selection, qact
            )
        elif qact == ACTION.QUEUE_BOTTOM:
            client.core.queue_bottom(self.torrent_ids).addCallback(
                self.move_selection, qact
            )
        elif qact == ACTION.QUEUE_UP:
            client.core.queue_up(self.torrent_ids).addCallback(
                self.move_selection, qact
            )
        elif qact == ACTION.QUEUE_DOWN:
            client.core.queue_down(self.torrent_ids).addCallback(
                self.move_selection, qact
            )

    def popup(self, **kwargs):
        popup = SelectablePopup(
            self.torrentslist,
            'Queue Action',
            self.do_queue,
            cb_args=kwargs,
            border_off_west=1,
        )
        popup.add_line(ACTION.QUEUE_TOP, '_Top')
        popup.add_line(ACTION.QUEUE_UP, '_Up')
        popup.add_line(ACTION.QUEUE_DOWN, '_Down')
        popup.add_line(ACTION.QUEUE_BOTTOM, '_Bottom')
        self.torrentslist.push_popup(popup)



#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import gi  # isort:skip (Required before Gtk import).

gi.require_version('Gtk', '3.0')

# isort:imports-thirdparty
from gi.repository import Gtk

# isort:imports-firstparty
import deluge.component as component
from deluge.ui.client import client

# isort:imports-localfolder
from ..common import get_resource

log = logging.getLogger(__name__)

NO_LABEL = 'No Label'


# menu
class LabelSidebarMenu:
    def __init__(self):

        self.treeview = component.get('FilterTreeView')
        self.menu = self.treeview.menu
        self.items = []

        # add items, in reverse order, because they are prepended.
        sep = Gtk.SeparatorMenuItem()
        self.items.append(sep)
        self.menu.prepend(sep)
        self._add_item('options', _('Label _Options'))
        self._add_item('remove', _('_Remove Label'))
        self._add_item('add', _('_Add Label'))

        self.menu.show_all()
        # dialogs:
        self.add_dialog = AddDialog()
        self.options_dialog = OptionsDialog()
        # hooks:
        self.menu.connect('show', self.on_show, None)

    def _add_item(self, item_id, label):
        """
        id is automatically-added as self.item_<id>
        """
        item = Gtk.MenuItem.new_with_mnemonic(label)
        func = getattr(self, 'on_%s' % item_id)
        item.connect('activate', func)
        self.menu.prepend(item)
        setattr(self, 'item_%s' % item_id, item)
        self.items.append(item)
        return item

    def on_add(self, event=None):
        self.add_dialog.show()

    def on_remove(self, event=None):
        client.label.remove(self.treeview.value)

    def on_options(self, event=None):
        self.options_dialog.show(self.treeview.value)

    def on_show(self, widget=None, data=None):
        """No Label:disable options/del."""
        log.debug('label-sidebar-popup:on-show')

        cat = self.treeview.cat
        label = self.treeview.value
        if cat == 'label' or (cat == 'cat' and label == 'label'):
            # is a label : show  menu-items
            for item in self.items:
                item.show()
            # default items
            sensitive = (label not in (NO_LABEL, None, '', 'All')) and (cat != 'cat')
            for item in self.items:
                item.set_sensitive(sensitive)

            # add is always enabled.
            self.item_add.set_sensitive(True)
        else:
            # not a label -->hide everything.
            for item in self.items:
                item.hide()

    def unload(self):
        log.debug('disable01')
        for item in list(self.items):
            item.hide()
            item.destroy()
            log.debug('disable02')
        self.items = []


# dialogs:
class AddDialog:
    def __init__(self):
        pass

    def show(self):
        self.builder = Gtk.Builder()
        self.builder.add_from_file(get_resource('label_add.ui'))
        self.dialog = self.builder.get_object('dlg_label_add')
        self.dialog.set_transient_for(component.get('MainWindow').window)

        self.builder.connect_signals(self)
        self.dialog.run()

    def on_add_ok(self, event=None):
        value = self.builder.get_object('txt_add').get_text()
        client.label.add(value)
        self.dialog.destroy()

    def on_add_cancel(self, event=None):
        self.dialog.destroy()


class OptionsDialog:
    spin_ids = ['max_download_speed', 'max_upload_speed', 'stop_ratio']
    spin_int_ids = ['max_upload_slots', 'max_connections']
    chk_ids = [
        'apply_max',
        'apply_queue',
        'stop_at_ratio',
        'apply_queue',
        'remove_at_ratio',
        'apply_move_completed',
        'move_completed',
        'is_auto_managed',
        'auto_add',
    ]

    # list of tuples, because order matters when nesting.
    sensitive_groups = [
        (
            'apply_max',
            [
                'max_download_speed',
                'max_upload_speed',
                'max_upload_slots',
                'max_connections',
            ],
        ),
        ('apply_queue', ['is_auto_managed', 'stop_at_ratio']),
        ('stop_at_ratio', ['remove_at_ratio', 'stop_ratio']),  # nested
        ('apply_move_completed', ['move_completed']),
        ('move_completed', ['move_completed_path']),  # nested
        ('auto_add', ['auto_add_trackers']),
    ]

    def __init__(self):
        pass

    def show(self, label):
        self.label = label
        self.builder = Gtk.Builder()
        self.builder.add_from_file(get_resource('label_options.ui'))
        self.dialog = self.builder.get_object('dlg_label_options')
        self.dialog.set_transient_for(component.get('MainWindow').window)
        self.builder.connect_signals(self)
        # Show the label name in the header label
        self.builder.get_object('label_header').set_markup(
            '<b>{}:</b> {}'.format(_('Label Options'), self.label)
        )

        for chk_id, group in self.sensitive_groups:
            chk = self.builder.get_object(chk_id)
            chk.connect('toggled', self.apply_sensitivity)

        client.label.get_options(self.label).addCallback(self.load_options)

        self.dialog.run()

    def load_options(self, options):
        log.debug(list(options))

        for spin_id in self.spin_ids + self.spin_int_ids:
            self.builder.get_object(spin_id).set_value(options[spin_id])
        for chk_id in self.chk_ids:
            self.builder.get_object(chk_id).set_active(bool(options[chk_id]))

        if client.is_localhost():
            self.builder.get_object('move_completed_path').set_filename(
                options['move_completed_path']
            )
            self.builder.get_object('move_completed_path').show()
            self.builder.get_object('move_completed_path_entry').hide()
        else:
            self.builder.get_object('move_completed_path_entry').set_text(
                options['move_completed_path']
            )
            self.builder.get_object('move_completed_path_entry').show()
            self.builder.get_object('move_completed_path').hide()

        self.builder.get_object('auto_add_trackers').get_buffer().set_text(
            '\n'.join(options['auto_add_trackers'])
        )

        self.apply_sensitivity()

    def on_options_ok(self, event=None):
        """Save options."""
        options = {}

        for spin_id in self.spin_ids:
            options[spin_id] = self.builder.get_object(spin_id).get_value()
        for spin_int_id in self.spin_int_ids:
            options[spin_int_id] = self.builder.get_object(
                spin_int_id
            ).get_value_as_int()
        for chk_id in self.chk_ids:
            options[chk_id] = self.builder.get_object(chk_id).get_active()

        if client.is_localhost():
            options['move_completed_path'] = self.builder.get_object(
                'move_completed_path'
            ).get_filename()
        else:
            options['move_completed_path'] = self.builder.get_object(
                'move_completed_path_entry'
            ).get_text()

        buff = self.builder.get_object(
            'auto_add_trackers'
        ).get_buffer()  # sometimes I hate gtk...
        tracker_lst = (
            buff.get_text(
                buff.get_start_iter(), buff.get_end_iter(), include_hidden_chars=False
            )
            .strip()
            .split('\n')
        )
        options['auto_add_trackers'] = [
            x for x in tracker_lst if x
        ]  # filter out empty lines.

        log.debug(options)
        client.label.set_options(self.label, options)
        self.dialog.destroy()

    def apply_sensitivity(self, event=None):
        for chk_id, sensitive_list in self.sensitive_groups:
            chk = self.builder.get_object(chk_id)
            sens = chk.get_active() and chk.get_property('sensitive')
            for widget_id in sensitive_list:
                self.builder.get_object(widget_id).set_sensitive(sens)

    def on_options_cancel(self, event=None):
        self.dialog.destroy()



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
import sys
import traceback

import pytest
from twisted.internet import defer, protocol, reactor
from twisted.internet.defer import Deferred
from twisted.internet.error import CannotListenError

import deluge.configmanager
import deluge.core.preferencesmanager
import deluge.log
from deluge.common import get_localhost_auth
from deluge.error import DelugeError
from deluge.ui.client import Client

# This sets log level to critical, so use log.critical() to debug while running unit tests
deluge.log.setup_logger('none')


def disable_new_release_check():
    deluge.core.preferencesmanager.DEFAULT_PREFS['new_release_check'] = False


def setup_test_logger(level='info', prefix='deluge'):
    deluge.log.setup_logger(level, filename='%s.log' % prefix, twisted_observer=False)


def get_test_data_file(filename):
    return os.path.join(os.path.join(os.path.dirname(__file__), 'data'), filename)


def todo_test(caller):
    # If we are using the delugereporter we can set todo mark on the test
    # Without the delugereporter the todo would print a stack trace, so in
    # that case we rely only on skipTest
    if os.environ.get('DELUGE_REPORTER', None):
        getattr(caller, caller._testMethodName).__func__.todo = 'To be fixed'

    filename = os.path.basename(traceback.extract_stack(None, 2)[0][0])
    funcname = traceback.extract_stack(None, 2)[0][2]
    pytest.skip(f'TODO: {filename}:{funcname}')


def add_watchdog(deferred, timeout=0.05, message=None):
    def callback(value):
        if not watchdog.called and not watchdog.cancelled:
            watchdog.cancel()
        if not deferred.called:
            if message:
                print(message)
            deferred.cancel()
        return value

    deferred.addBoth(callback)
    watchdog = reactor.callLater(timeout, defer.Deferred.addTimeout, deferred)
    return watchdog


class ReactorOverride:
    """Class used to patch reactor while running unit tests
    to avoid starting and stopping the twisted reactor
    """

    def __getattr__(self, attr):
        if attr == 'run':
            return self._run
        if attr == 'stop':
            return self._stop
        return getattr(reactor, attr)

    def _run(self):
        pass

    def _stop(self):
        pass

    def addReader(self, arg):  # NOQA: N802
        pass


class ProcessOutputHandler(protocol.ProcessProtocol):
    def __init__(
        self,
        script,
        shutdown_func,
        callbacks,
        logfile=None,
        print_stdout=True,
        print_stderr=True,
    ):
        """Executes a script and handle the process' output to stdout and stderr.

        Args:
            script (str): The script to execute.
            shutdown_func (func): A function which will gracefully stop the called script.
            callbacks (list): Callbacks to trigger if the expected output if found.
            logfile (str, optional): Filename to wrote the process' output.
            print_stderr (bool): Print the process' stderr output to stdout.
            print_stdout (bool): Print the process' stdout output to stdout.

        """
        self.callbacks = callbacks
        self.script = script
        self.shutdown_func = shutdown_func
        self.log_output = ''
        self.stderr_out = ''
        self.logfile = logfile
        self.print_stdout = print_stdout
        self.print_stderr = print_stderr
        self.quit_d = None
        self.killed = False
        self.watchdogs = []

    def connectionMade(self):  # NOQA: N802
        self.transport.write(self.script)
        self.transport.closeStdin()

    def outConnectionLost(self):  # NOQA: N802
        if not self.logfile:
            return
        with open(self.logfile, 'w') as f:
            f.write(self.log_output)

    @defer.inlineCallbacks
    def kill(self):
        """Kill the running process.

        Returns:
            Deferred: A deferred that is triggered when the process has quit.

        """
        if self.killed:
            return
        self.killed = True
        self._kill_watchdogs()
        self.quit_d = Deferred()
        shutdown = self.shutdown_func()
        shutdown.addTimeout(5, reactor)
        try:
            yield shutdown
        except Exception:
            self.transport.signalProcess('TERM')
        result = yield self.quit_d
        return result

    def _kill_watchdogs(self):
        """Cancel all watchdogs"""
        for w in self.watchdogs:
            if not w.called and not w.cancelled:
                w.cancel()

    def processEnded(self, status):  # NOQA: N802
        self.transport.loseConnection()
        if self.quit_d is None:
            return
        if status.value.exitCode == 0:
            self.quit_d.callback(True)
        else:
            self.quit_d.errback(status)

    def check_callbacks(self, data, cb_type='stdout'):
        ret = False
        for c in self.callbacks:
            if cb_type not in c['types'] or c['deferred'].called:
                continue
            for trigger in c['triggers']:
                if trigger['expr'] in data:
                    ret = True
                    if 'cb' in trigger:
                        trigger['cb'](self, c['deferred'], data, self.log_output)
                    elif 'value' not in trigger:
                        raise Exception('Trigger must specify either "cb" or "value"')
                    else:
                        val = trigger['value'](self, data, self.log_output)
                        if trigger.get('type', 'callback') == 'errback':
                            c['deferred'].errback(val)
                        else:
                            c['deferred'].callback(val)
        return ret

    def outReceived(self, data):  # NOQA: N802
        """Process output from stdout"""
        data = data.decode('utf8')
        self.log_output += data
        if self.check_callbacks(data):
            pass
        elif '[ERROR' in data:
            if not self.print_stdout:
                return
            print(data, end=' ')

    def errReceived(self, data):  # NOQA: N802
        """Process output from stderr"""
        data = data.decode('utf8')
        self.log_output += data
        self.stderr_out += data
        self.check_callbacks(data, cb_type='stderr')
        if not self.print_stderr:
            return
        data = '\n%s' % data.strip()
        prefixed = data.replace('\n', '\nSTDERR: ')
        print('\n%s' % prefixed)


def start_core(
    listen_port=58900,
    logfile=None,
    timeout=10,
    timeout_msg=None,
    custom_script='',
    print_stdout=True,
    print_stderr=True,
    extra_callbacks=None,
    config_directory='',
):
    """Start the deluge core as a daemon.

    Args:
        listen_port (int, optional): The port the daemon listens for client connections.
        logfile (str, optional): Logfile name to write the output from the process.
        timeout (int): If none of the callbacks have been triggered before the timeout, the process is killed.
        timeout_msg (str): The message to print when the timeout expires.
        custom_script (str): Extra python code to insert into the daemon process script.
        print_stderr (bool): If the output from the process' stderr should be printed to stdout.
        print_stdout (bool): If the output from the process' stdout should be printed to stdout.
        extra_callbacks (list): A list of dictionaries specifying extra callbacks.

    Returns:
        tuple(Deferred, ProcessOutputHandler):

        The Deferred is fired when the core callback is triggered either after the default
        output triggers are matched (daemon successfully started, or failed to start),
        or upon timeout expiry. The ProcessOutputHandler is the handler for the deluged process.

    """
    daemon_script = """
import sys
import deluge.core.daemon_entry

from deluge.common import windows_check

if windows_check():
    sys.argv.extend(['-c', '%(dir)s', '-L', 'info', '-p', '%(port)d'])
else:
    sys.argv.extend(['-d', '-c', '%(dir)s', '-L', 'info', '-p', '%(port)d'])

try:
    daemon = deluge.core.daemon_entry.start_daemon(skip_start=True)
    %(script)s
    daemon.start()
except Exception:
    import traceback
    sys.stderr.write('Exception raised:\\n %%s' %% traceback.format_exc())
""" % {
        'dir': config_directory.as_posix(),
        'port': listen_port,
        'script': custom_script,
    }

    callbacks = []
    default_core_cb = {'deferred': Deferred(), 'types': 'stdout'}
    if timeout:
        default_core_cb['timeout'] = timeout

    # Specify the triggers for daemon log output
    default_core_cb['triggers'] = [
        {'expr': 'Finished loading ', 'value': lambda reader, data, data_all: reader},
        {
            'expr': 'Cannot start deluged, listen port in use.',
            'type': 'errback',
            'value': lambda reader, data, data_all: CannotListenError(
                'localhost',
                listen_port,
                'Could not start deluge test client!\n%s' % data,
            ),
        },
        {
            'expr': 'Traceback',
            'type': 'errback',
            'value': lambda reader, data, data_all: DelugeError(
                'Traceback found when starting daemon:\n%s' % data
            ),
        },
    ]

    callbacks.append(default_core_cb)
    if extra_callbacks:
        callbacks.extend(extra_callbacks)

    @defer.inlineCallbacks
    def shutdown_daemon():
        username, password = get_localhost_auth()
        client = Client()
        yield client.connect(
            'localhost', listen_port, username=username, password=password
        )
        yield client.daemon.shutdown()

    process_protocol = start_process(
        daemon_script, shutdown_daemon, callbacks, logfile, print_stdout, print_stderr
    )
    return default_core_cb['deferred'], process_protocol


def start_process(
    script, shutdown_func, callbacks, logfile=None, print_stdout=True, print_stderr=True
):
    """
    Starts an external python process which executes the given script.

    Args:
        script (str): The content of the script to execute.
        shutdown_func (func): A function which will gracefully end the called script.
        callbacks (list): list of dictionaries specifying callbacks.
        logfile (str, optional): Logfile name to write the output from the process.
        print_stderr (bool): If the output from the process' stderr should be printed to stdout.
        print_stdout (bool): If the output from the process' stdout should be printed to stdout.

    Returns:
        ProcessOutputHandler: The handler for the process's output.

    Each entry in the callbacks list is a dictionary with the following keys:
      * "deferred": The deferred to be called when matched.
      * "types": The output this callback should be matched against.
                 Possible values: ["stdout", "stderr"]
      * "timeout" (optional): A timeout in seconds for the deferred.
      * "triggers": A list of dictionaries, each specifying specifying a trigger:
        * "expr": A string to match against the log output.
        * "value": A function to produce the result to be passed to the callback.
        * "type" (optional): A string that specifies wether to trigger a regular callback or errback.

    """
    cwd = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    process_protocol = ProcessOutputHandler(
        script.encode('utf8'),
        shutdown_func,
        callbacks,
        logfile,
        print_stdout,
        print_stderr,
    )

    # Add timeouts to deferreds
    for c in callbacks:
        if 'timeout' in c:
            w = add_watchdog(
                c['deferred'], timeout=c['timeout'], message=c.get('timeout_msg', None)
            )
            process_protocol.watchdogs.append(w)

    reactor.spawnProcess(
        process_protocol, sys.executable, args=[sys.executable], path=cwd
    )
    return process_protocol



#
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from gi.repository import Gdk, Gtk

import deluge.component as component
from deluge.plugins.pluginbase import Gtk3PluginBase
from deluge.ui.client import client

from .common import get_resource

log = logging.getLogger(__name__)

DAYS = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']


class SchedulerSelectWidget(Gtk.DrawingArea):
    def __init__(self, hover):
        super().__init__()
        self.set_events(
            Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
            | Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.LEAVE_NOTIFY_MASK
        )

        self.connect('draw', self.draw)
        self.connect('button_press_event', self.mouse_down)
        self.connect('button_release_event', self.mouse_up)
        self.connect('motion_notify_event', self.mouse_hover)
        self.connect('leave_notify_event', self.mouse_leave)

        self.colors = [
            [115 / 255, 210 / 255, 22 / 255],
            [237 / 255, 212 / 255, 0 / 255],
            [204 / 255, 0 / 255, 0 / 255],
        ]
        self.button_state = [[0] * 7 for dummy in range(24)]

        self.start_point = [0, 0]
        self.hover_point = [-1, -1]
        self.hover_label = hover
        self.hover_days = DAYS
        self.mouse_press = False
        self.set_size_request(350, 150)

    def set_button_state(self, state):
        self.button_state = []
        for s in state:
            self.button_state.append(list(s))
        log.debug(self.button_state)

    # redraw the whole thing
    def draw(self, widget, context):
        width = widget.get_allocated_width()
        height = widget.get_allocated_height()
        context.rectangle(0, 0, width, height)
        context.clip()

        for y in range(7):
            for x in range(24):
                context.set_source_rgba(
                    self.colors[self.button_state[x][y]][0],
                    self.colors[self.button_state[x][y]][1],
                    self.colors[self.button_state[x][y]][2],
                    0.5,
                )
                context.rectangle(
                    width * (6 * x / 145 + 1 / 145),
                    height * (6 * y / 43 + 1 / 43),
                    6 * width / 145,
                    5 * height / 43,
                )
                context.fill_preserve()
                context.set_source_rgba(0, 0, 0, 0.7)
                context.set_line_width(1)
                context.stroke()

    # coordinates --> which box
    def get_point(self, event):
        width = self.get_allocated_width()
        height = self.get_allocated_height()
        x = int((event.x - width * 0.5 / 145) / (6 * width / 145))
        y = int((event.y - height * 0.5 / 43) / (6 * height / 43))

        if x > 23:
            x = 23
        elif x < 0:
            x = 0
        if y > 6:
            y = 6
        elif y < 0:
            y = 0

        return [x, y]

    # mouse down
    def mouse_down(self, widget, event):
        self.mouse_press = True
        self.start_point = self.get_point(event)

    # if the same box -> change it
    def mouse_up(self, widget, event):
        self.mouse_press = False
        end_point = self.get_point(event)

        # change color on mouseclick depending on the button
        if end_point[0] is self.start_point[0] and end_point[1] is self.start_point[1]:
            if event.button == 1:
                self.button_state[end_point[0]][end_point[1]] += 1
                if self.button_state[end_point[0]][end_point[1]] > 2:
                    self.button_state[end_point[0]][end_point[1]] = 0
            elif event.button == 3:
                self.button_state[end_point[0]][end_point[1]] -= 1
                if self.button_state[end_point[0]][end_point[1]] < 0:
                    self.button_state[end_point[0]][end_point[1]] = 2
            self.queue_draw()

    # if box changed and mouse is pressed draw all boxes from start point to end point
    # set hover text etc..
    def mouse_hover(self, widget, event):
        if self.get_point(event) != self.hover_point:
            self.hover_point = self.get_point(event)

            self.hover_label.set_text(
                self.hover_days[self.hover_point[1]]
                + ' '
                + str(self.hover_point[0])
                + ':00 - '
                + str(self.hover_point[0])
                + ':59'
            )

            if self.mouse_press:
                points = [
                    [self.hover_point[0], self.start_point[0]],
                    [self.hover_point[1], self.start_point[1]],
                ]

                for x in range(min(points[0]), max(points[0]) + 1):
                    for y in range(min(points[1]), max(points[1]) + 1):
                        self.button_state[x][y] = self.button_state[
                            self.start_point[0]
                        ][self.start_point[1]]

                self.queue_draw()

    # clear hover text on mouse leave
    def mouse_leave(self, widget, event):
        self.hover_label.set_text('')
        self.hover_point = [-1, -1]


class GtkUI(Gtk3PluginBase):
    def enable(self):
        self.create_prefs_page()

        component.get('PluginManager').register_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').register_hook(
            'on_show_prefs', self.on_show_prefs
        )
        self.statusbar = component.get('StatusBar')
        self.status_item = self.statusbar.add_item(
            image=get_resource('green.svg'),
            text='',
            callback=self.on_status_item_clicked,
            tooltip='Scheduler',
        )

        def on_state_deferred(state):
            self.state = state
            self.on_scheduler_event(state)

        self.on_show_prefs()

        client.scheduler.get_state().addCallback(on_state_deferred)
        client.register_event_handler('SchedulerEvent', self.on_scheduler_event)

    def disable(self):
        component.get('Preferences').remove_page(_('Scheduler'))
        # Reset statusbar dict.
        self.statusbar.config_value_changed_dict[
            'max_download_speed'
        ] = self.statusbar._on_max_download_speed
        self.statusbar.config_value_changed_dict[
            'max_upload_speed'
        ] = self.statusbar._on_max_upload_speed
        # Remove statusbar item.
        self.statusbar.remove_item(self.status_item)
        del self.status_item

        component.get('PluginManager').deregister_hook(
            'on_apply_prefs', self.on_apply_prefs
        )
        component.get('PluginManager').deregister_hook(
            'on_show_prefs', self.on_show_prefs
        )

    def on_apply_prefs(self):
        log.debug('applying prefs for Scheduler')
        config = {}
        config['low_down'] = self.spin_download.get_value()
        config['low_up'] = self.spin_upload.get_value()
        config['low_active'] = self.spin_active.get_value_as_int()
        config['low_active_down'] = self.spin_active_down.get_value_as_int()
        config['low_active_up'] = self.spin_active_up.get_value_as_int()
        config['button_state'] = self.scheduler_select.button_state
        client.scheduler.set_config(config)

    def on_show_prefs(self):
        def on_get_config(config):
            log.debug('config: %s', config)
            self.scheduler_select.set_button_state(config['button_state'])
            self.spin_download.set_value(config['low_down'])
            self.spin_upload.set_value(config['low_up'])
            self.spin_active.set_value(config['low_active'])
            self.spin_active_down.set_value(config['low_active_down'])
            self.spin_active_up.set_value(config['low_active_up'])

        client.scheduler.get_config().addCallback(on_get_config)

    def on_scheduler_event(self, state):
        self.state = state
        self.status_item.set_image_from_file(get_resource(self.state.lower() + '.svg'))
        if self.state == 'Yellow':
            # Prevent func calls in Statusbar if the config changes.
            self.statusbar.config_value_changed_dict.pop('max_download_speed', None)
            self.statusbar.config_value_changed_dict.pop('max_upload_speed', None)
            try:
                self.statusbar._on_max_download_speed(self.spin_download.get_value())
                self.statusbar._on_max_upload_speed(self.spin_upload.get_value())
            except AttributeError:
                # Skip error due to Plugin being enabled before statusbar items created on startup.
                pass
        else:
            self.statusbar.config_value_changed_dict[
                'max_download_speed'
            ] = self.statusbar._on_max_download_speed
            self.statusbar.config_value_changed_dict[
                'max_upload_speed'
            ] = self.statusbar._on_max_upload_speed

            def update_config_values(config):
                try:
                    self.statusbar._on_max_download_speed(config['max_download_speed'])
                    self.statusbar._on_max_upload_speed(config['max_upload_speed'])
                except AttributeError:
                    # Skip error due to Plugin being enabled before statusbar items created on startup.
                    pass

            client.core.get_config_values(
                ['max_download_speed', 'max_upload_speed']
            ).addCallback(update_config_values)

    def on_status_item_clicked(self, widget, event):
        component.get('Preferences').show('Scheduler')

    # Configuration dialog
    def create_prefs_page(self):
        # Select Widget
        hover = Gtk.Label()
        self.scheduler_select = SchedulerSelectWidget(hover)

        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, spacing=5)
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=5)
        vbox_days = Gtk.Box.new(Gtk.Orientation.VERTICAL, spacing=0)
        for day in DAYS:
            vbox_days.pack_start(Gtk.Label(day, xalign=0), True, False, 0)
        hbox.pack_start(vbox_days, False, False, 15)
        hbox.pack_start(self.scheduler_select, True, True, 0)
        frame = Gtk.Frame()
        label = Gtk.Label()
        label.set_markup(_('<b>Schedule</b>'))
        frame.set_label_widget(label)
        frame.set_shadow_type(Gtk.ShadowType.NONE)
        frame.set_margin_left(15)
        frame.add(hbox)

        vbox.pack_start(frame, False, False, 0)
        vbox.pack_start(hover, False, False, 0)

        table = Gtk.Table(5, 2)
        table.set_margin_left(15)

        label = Gtk.Label(_('Download Limit:'))
        label.set_alignment(0.0, 0.6)
        table.attach_defaults(label, 0, 1, 0, 1)
        self.spin_download = Gtk.SpinButton()
        self.spin_download.set_numeric(True)
        self.spin_download.set_range(-1.0, 99999.0)
        self.spin_download.set_increments(1, 10)
        table.attach_defaults(self.spin_download, 1, 2, 0, 1)

        label = Gtk.Label(_('Upload Limit:'))
        label.set_alignment(0.0, 0.6)
        table.attach_defaults(label, 0, 1, 1, 2)
        self.spin_upload = Gtk.SpinButton()
        self.spin_upload.set_numeric(True)
        self.spin_upload.set_range(-1.0, 99999.0)
        self.spin_upload.set_increments(1, 10)
        table.attach_defaults(self.spin_upload, 1, 2, 1, 2)

        label = Gtk.Label(_('Active Torrents:'))
        label.set_alignment(0.0, 0.6)
        table.attach_defaults(label, 0, 1, 2, 3)
        self.spin_active = Gtk.SpinButton()
        self.spin_active.set_numeric(True)
        self.spin_active.set_range(-1, 9999)
        self.spin_active.set_increments(1, 10)
        table.attach_defaults(self.spin_active, 1, 2, 2, 3)

        label = Gtk.Label(_('Active Downloading:'))
        label.set_alignment(0.0, 0.6)
        table.attach_defaults(label, 0, 1, 3, 4)
        self.spin_active_down = Gtk.SpinButton()
        self.spin_active_down.set_numeric(True)
        self.spin_active_down.set_range(-1, 9999)
        self.spin_active_down.set_increments(1, 10)
        table.attach_defaults(self.spin_active_down, 1, 2, 3, 4)

        label = Gtk.Label(_('Active Seeding:'))
        label.set_alignment(0.0, 0.6)
        table.attach_defaults(label, 0, 1, 4, 5)
        self.spin_active_up = Gtk.SpinButton()
        self.spin_active_up.set_numeric(True)
        self.spin_active_up.set_range(-1, 9999)
        self.spin_active_up.set_increments(1, 10)
        table.attach_defaults(self.spin_active_up, 1, 2, 4, 5)

        eventbox = Gtk.EventBox()
        eventbox.add(table)
        frame = Gtk.Frame()
        label = Gtk.Label()
        label.set_markup(_('<b>Slow Settings</b>'))
        label.modify_bg(Gtk.StateFlags.NORMAL, Gdk.color_parse('#EDD400'))
        frame.set_label_widget(label)
        frame.set_margin_left(15)
        frame.set_border_width(2)
        frame.add(eventbox)
        vbox.pack_start(frame, False, False, 0)

        vbox.show_all()
        component.get('Preferences').add_page(_('Scheduler'), vbox)



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import deluge.component as component
from deluge.common import get_localhost_auth
from deluge.conftest import BaseTestCase
from deluge.core.authmanager import AUTH_LEVEL_ADMIN, AuthManager


class TestAuthManager(BaseTestCase):
    def set_up(self):
        self.auth = AuthManager()
        self.auth.start()

    def tear_down(self):
        # We must ensure that the components in component registry are removed
        return component.shutdown()

    def test_authorize(self):
        assert self.auth.authorize(*get_localhost_auth()) == AUTH_LEVEL_ADMIN



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os.path

import pytest
from twisted.internet import defer
from twisted.internet.error import CannotListenError

import deluge.component as component

from . import common


@pytest.mark.usefixtures('config_dir')
class DaemonBase:
    def common_set_up(self):
        self.listen_port = 58900
        self.core = None
        return component.start()

    def terminate_core(self, *args):
        if args[0] is not None:
            if hasattr(args[0], 'getTraceback'):
                print('terminate_core: Errback Exception: %s' % args[0].getTraceback())

        if not self.core.killed:
            d = self.core.kill()
            return d

    @defer.inlineCallbacks
    def start_core(
        self,
        arg,
        custom_script='',
        logfile='',
        print_stdout=True,
        print_stderr=True,
        timeout=5,
        port_range=10,
        extra_callbacks=None,
    ):
        if logfile == '':
            logfile = 'daemon_%s.log' % self.id()

        # We are running py.test
        if hasattr(pytest, 'config'):
            if self.basetemp:
                if not os.path.exists(self.basetemp):
                    os.makedirs(self.basetemp)
                logfile = os.path.join(self.basetemp, logfile)

        for dummy in range(port_range):
            try:
                d, self.core = common.start_core(
                    listen_port=self.listen_port,
                    logfile=logfile,
                    timeout=timeout,
                    timeout_msg='Timeout!',
                    custom_script=custom_script,
                    print_stdout=print_stdout,
                    print_stderr=print_stderr,
                    extra_callbacks=extra_callbacks,
                    config_directory=self.config_dir,
                )
                yield d
            except CannotListenError as ex:
                exception_error = ex
                self.listen_port += 1
            except (KeyboardInterrupt, SystemExit):
                raise
            else:
                return
        raise exception_error



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import pytest
import pytest_twisted
from twisted.internet import defer

import deluge.component as component
from deluge.common import fsize, fspeed
from deluge.ui.client import client


def print_totals(totals):
    for name, value in totals.items():
        print(name, fsize(value))

    print('overhead:')
    print('up:', fsize(totals['total_upload'] - totals['total_payload_upload']))
    print('down:', fsize(totals['total_download'] - totals['total_payload_download']))


@pytest.mark.usefixtures('component')
class TestStatsPlugin:
    @pytest_twisted.async_yield_fixture(autouse=True)
    async def set_up(self):
        defer.setDebugging(True)
        client.start_standalone()
        client.core.enable_plugin('Stats')
        await component.start()
        yield
        client.stop_standalone()
        await component.shutdown()

    @defer.inlineCallbacks
    def test_client_totals(self):
        plugins = yield client.core.get_available_plugins()
        if 'Stats' not in plugins:
            pytest.skip('Stats plugin not available for testing')

        totals = yield client.stats.get_totals()
        assert totals['total_upload'] == 0
        assert totals['total_payload_upload'] == 0
        assert totals['total_payload_download'] == 0
        assert totals['total_download'] == 0
        # print_totals(totals)

    @defer.inlineCallbacks
    def test_session_totals(self):
        plugins = yield client.core.get_available_plugins()
        if 'Stats' not in plugins:
            pytest.skip('Stats plugin not available for testing')

        totals = yield client.stats.get_session_totals()
        assert totals['total_upload'] == 0
        assert totals['total_payload_upload'] == 0
        assert totals['total_payload_download'] == 0
        assert totals['total_download'] == 0
        # print_totals(totals)

    @pytest.mark.gtkui
    @defer.inlineCallbacks
    def test_write(self):
        """
        writing to a file-like object; need this for webui.

        Not strictly a unit test, but tests if calls do not fail...
        """
        from deluge_stats import graph, gtkui

        from deluge.configmanager import ConfigManager
        from deluge.ui.gtk3.gtkui import DEFAULT_PREFS
        from deluge.ui.gtk3.mainwindow import MainWindow
        from deluge.ui.gtk3.pluginmanager import PluginManager
        from deluge.ui.gtk3.preferences import Preferences
        from deluge.ui.gtk3.torrentdetails import TorrentDetails
        from deluge.ui.gtk3.torrentview import TorrentView

        ConfigManager('gtk3ui.conf', defaults=DEFAULT_PREFS)

        self.plugins = PluginManager()
        MainWindow()
        TorrentView()
        TorrentDetails()
        Preferences()

        class FakeFile:
            def __init__(self):
                self.data = []

            def write(self, data):
                self.data.append(data)

        stats_gtkui = gtkui.GtkUI('test_stats')
        stats_gtkui.enable()
        yield stats_gtkui.graphs_tab.update()

        g = stats_gtkui.graphs_tab.graph
        g.add_stat('download_rate', color=graph.green)
        g.add_stat('upload_rate', color=graph.blue)
        g.set_left_axis(formatter=fspeed, min=10240)

        surface = g.draw(900, 150)
        file_like = FakeFile()
        surface.write_to_png(file_like)
        data = b''.join(file_like.data)
        with open('file_like.png', 'wb') as _file:
            _file.write(data)



#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

from gi.repository.Gtk import Builder

from deluge.ui.client import client

from ..common import get_resource

log = logging.getLogger(__name__)


class LabelConfig:
    """
    there used to be some options here...
    """

    def __init__(self, plugin):
        self.plugin = plugin

    def load(self):
        log.debug('Adding Label Preferences page')
        builder = Builder()
        builder.add_from_file(get_resource('label_pref.ui'))

        self.plugin.add_preferences_page(
            _('Label'), builder.get_object('label_prefs_box')
        )
        self.plugin.register_hook('on_show_prefs', self.load_settings)
        self.plugin.register_hook('on_apply_prefs', self.on_apply_prefs)

        self.load_settings()

    def unload(self):
        self.plugin.remove_preferences_page(_('Label'))
        self.plugin.deregister_hook('on_apply_prefs', self.on_apply_prefs)
        self.plugin.deregister_hook('on_show_prefs', self.load_settings)

    def load_settings(self, widget=None, data=None):
        client.label.get_config().addCallback(self.cb_global_options)

    def cb_global_options(self, options):
        log.debug('options=%s', options)

        # for id in self.chk_ids:
        #    self.glade.get_widget(id).set_active(bool(options[id]))

    def on_apply_prefs(self):
        options = {}
        # update options dict here.
        client.label.set_config(options)



#
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import gi  # isort:skip (Required before Gtk import).

gi.require_version('Gtk', '3.0')

# isort:imports-thirdparty
from gi.repository import Gtk

# isort:imports-firstparty
import deluge.component as component
from deluge.plugins.pluginbase import Gtk3PluginBase
from deluge.ui.client import client

# isort:imports-localfolder
from . import common

log = logging.getLogger(__name__)

EXECUTE_ID = 0
EXECUTE_EVENT = 1
EXECUTE_COMMAND = 2

EVENT_MAP = {
    'complete': _('Torrent Complete'),
    'added': _('Torrent Added'),
    'removed': _('Torrent Removed'),
}

EVENTS = ['complete', 'added', 'removed']


class ExecutePreferences:
    def __init__(self, plugin):
        self.plugin = plugin

    def load(self):
        log.debug('Adding Execute Preferences page')
        self.builder = Gtk.Builder()
        self.builder.add_from_file(common.get_resource('execute_prefs.ui'))
        self.builder.connect_signals(self)

        events = self.builder.get_object('event_combobox')

        store = Gtk.ListStore(str, str)
        for event in EVENTS:
            event_label = EVENT_MAP[event]
            store.append((event_label, event))
        events.set_model(store)
        events.set_active(0)

        self.plugin.add_preferences_page(
            _('Execute'), self.builder.get_object('execute_box')
        )
        self.plugin.register_hook('on_show_prefs', self.load_commands)
        self.plugin.register_hook('on_apply_prefs', self.on_apply_prefs)

        self.load_commands()

        client.register_event_handler(
            'ExecuteCommandAddedEvent', self.on_command_added_event
        )
        client.register_event_handler(
            'ExecuteCommandRemovedEvent', self.on_command_removed_event
        )

    def unload(self):
        self.plugin.remove_preferences_page(_('Execute'))
        self.plugin.deregister_hook('on_apply_prefs', self.on_apply_prefs)
        self.plugin.deregister_hook('on_show_prefs', self.load_commands)

    def add_command(self, command_id, event, command):
        log.debug('Adding command `%s`', command_id)
        vbox = self.builder.get_object('commands_vbox')
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=5)
        hbox.set_name(command_id + '_' + event)
        label = Gtk.Label(EVENT_MAP[event])
        entry = Gtk.Entry()
        entry.set_text(command)
        button = Gtk.Button()
        button.set_name('remove_%s' % command_id)
        button.connect('clicked', self.on_remove_button_clicked)

        img = Gtk.Image()
        img.set_from_stock(Gtk.STOCK_REMOVE, Gtk.IconSize.BUTTON)
        button.set_image(img)

        hbox.pack_start(label, False, False, 0)
        hbox.pack_start(entry, False, False, 0)
        hbox.pack_start(button, True, True, 0)
        hbox.show_all()
        vbox.pack_start(hbox, True, True, 0)

    def remove_command(self, command_id):
        vbox = self.builder.get_object('commands_vbox')
        children = vbox.get_children()
        for child in children:
            if child.get_name().split('_')[0] == command_id:
                vbox.remove(child)
                break

    def clear_commands(self):
        vbox = self.builder.get_object('commands_vbox')
        children = vbox.get_children()
        for child in children:
            vbox.remove(child)

    def load_commands(self):
        def on_get_commands(commands):
            self.clear_commands()
            log.debug('on_get_commands: %s', commands)
            for command in commands:
                command_id, event, command = command
                self.add_command(command_id, event, command)

        client.execute.get_commands().addCallback(on_get_commands)

    def on_add_button_clicked(self, *args):
        command = self.builder.get_object('command_entry').get_text()
        events = self.builder.get_object('event_combobox')
        event = events.get_model()[events.get_active()][1]
        client.execute.add_command(event, command)

    def on_remove_button_clicked(self, widget, *args):
        command_id = widget.get_name().replace('remove_', '')
        client.execute.remove_command(command_id)

    def on_apply_prefs(self):
        vbox = self.builder.get_object('commands_vbox')
        children = vbox.get_children()
        for child in children:
            command_id, event = child.get_name().split('_')
            for widget in child.get_children():
                if isinstance(widget, Gtk.Entry):
                    command = widget.get_text()
            client.execute.save_command(command_id, event, command)

    def on_command_added_event(self, command_id, event, command):
        log.debug('Adding command %s: %s', event, command)
        self.add_command(command_id, event, command)

    def on_command_removed_event(self, command_id):
        log.debug('Removing command %s', command_id)
        self.remove_command(command_id)


class GtkUI(Gtk3PluginBase):
    def enable(self):
        self.plugin = component.get('PluginManager')
        self.preferences = ExecutePreferences(self.plugin)
        self.preferences.load()

    def disable(self):
        self.preferences.unload()



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import json
from io import BytesIO

import pytest
import pytest_twisted
from twisted.internet import defer, reactor
from twisted.web.client import Agent, FileBodyProducer
from twisted.web.http_headers import Headers
from twisted.web.static import File

import deluge.component as component

from . import common
from .common_web import WebServerTestBase

common.disable_new_release_check()


class TestWebAPI(WebServerTestBase):
    @pytest.mark.xfail(reason='This just logs an error at the moment.')
    @pytest_twisted.ensureDeferred
    async def test_connect_invalid_host(self):
        with pytest.raises(Exception):
            await self.deluge_web.web_api.connect('id')

    def test_connect(self, client):
        d = self.deluge_web.web_api.connect(self.host_id)

        def on_connect(result):
            assert type(result) == tuple
            assert len(result) > 0
            return result

        d.addCallback(on_connect)
        d.addErrback(self.fail)
        return d

    def test_disconnect(self):
        d = self.deluge_web.web_api.connect(self.host_id)

        @defer.inlineCallbacks
        def on_connect(result):
            assert self.deluge_web.web_api.connected()
            yield self.deluge_web.web_api.disconnect()
            assert not self.deluge_web.web_api.connected()

        d.addCallback(on_connect)
        d.addErrback(self.fail)
        return d

    def test_get_config(self):
        config = self.deluge_web.web_api.get_config()
        assert self.webserver_listen_port == config['port']

    def test_set_config(self):
        config = self.deluge_web.web_api.get_config()
        config['pwd_salt'] = 'new_salt'
        config['pwd_sha1'] = 'new_sha'
        config['sessions'] = {
            '233f23632af0a74748bc5dd1d8717564748877baa16420e6898e17e8aa365e6e': {
                'login': 'skrot',
                'expires': 1460030877.0,
                'level': 10,
            }
        }
        self.deluge_web.web_api.set_config(config)
        web_config = component.get('DelugeWeb').config.config
        assert config['pwd_salt'] != web_config['pwd_salt']
        assert config['pwd_sha1'] != web_config['pwd_sha1']
        assert config['sessions'] != web_config['sessions']

    @defer.inlineCallbacks
    def get_host_status(self):
        host = list(self.deluge_web.web_api._get_host(self.host_id))
        host[3] = 'Online'
        host[4] = '2.0.0.dev562'
        status = yield self.deluge_web.web_api.get_host_status(self.host_id)
        assert status == tuple(status)

    def test_get_host(self):
        assert not self.deluge_web.web_api._get_host('invalid_id')
        conn = list(self.deluge_web.web_api.hostlist.get_hosts_info()[0])
        assert self.deluge_web.web_api._get_host(conn[0]) == conn[0:4]

    def test_add_host(self):
        conn = ['abcdef', '10.0.0.1', 0, 'user123', 'pass123']
        assert not self.deluge_web.web_api._get_host(conn[0])
        # Add valid host
        result, host_id = self.deluge_web.web_api.add_host(
            conn[1], conn[2], conn[3], conn[4]
        )
        assert result
        conn[0] = host_id
        assert self.deluge_web.web_api._get_host(conn[0]) == conn[0:4]

        # Add already existing host
        ret = self.deluge_web.web_api.add_host(conn[1], conn[2], conn[3], conn[4])
        assert ret == (False, 'Host details already in hostlist')

        # Add invalid port
        conn[2] = 'bad port'
        ret = self.deluge_web.web_api.add_host(conn[1], conn[2], conn[3], conn[4])
        assert ret == (False, 'Invalid port. Must be an integer')

    def test_remove_host(self):
        conn = ['connection_id', '', 0, '', '']
        self.deluge_web.web_api.hostlist.config['hosts'].append(conn)
        assert self.deluge_web.web_api._get_host(conn[0]) == conn[0:4]
        # Remove valid host
        assert self.deluge_web.web_api.remove_host(conn[0])
        assert not self.deluge_web.web_api._get_host(conn[0])
        # Remove non-existing host
        assert not self.deluge_web.web_api.remove_host(conn[0])

    def test_get_torrent_info(self):
        filename = common.get_test_data_file('test.torrent')
        ret = self.deluge_web.web_api.get_torrent_info(filename)
        assert ret['name'] == 'azcvsupdater_2.6.2.jar'
        assert ret['info_hash'] == 'ab570cdd5a17ea1b61e970bb72047de141bce173'
        assert 'files_tree' in ret

    def test_get_torrent_info_with_md5(self):
        filename = common.get_test_data_file('md5sum.torrent')
        ret = self.deluge_web.web_api.get_torrent_info(filename)
        # JSON dumping happens during response creation in normal usage
        # JSON serialization may fail if any of the dictionary items are byte arrays rather than strings
        ret = json.loads(json.dumps(ret))
        assert ret['name'] == 'test'
        assert ret['info_hash'] == 'f6408ba9944cf9fe01b547b28f336b3ee6ec32c5'
        assert 'files_tree' in ret

    def test_get_magnet_info(self):
        ret = self.deluge_web.web_api.get_magnet_info(
            'magnet:?xt=urn:btih:SU5225URMTUEQLDXQWRB2EQWN6KLTYKN'
        )
        assert ret['name'] == '953bad769164e8482c7785a21d12166f94b9e14d'
        assert ret['info_hash'] == '953bad769164e8482c7785a21d12166f94b9e14d'
        assert 'files_tree' in ret

    @pytest_twisted.inlineCallbacks
    def test_get_torrent_files(self):
        yield self.deluge_web.web_api.connect(self.host_id)
        filename = common.get_test_data_file('test.torrent')
        torrents = [
            {'path': filename, 'options': {'download_location': '/home/deluge/'}}
        ]
        yield self.deluge_web.web_api.add_torrents(torrents)
        ret = yield self.deluge_web.web_api.get_torrent_files(
            'ab570cdd5a17ea1b61e970bb72047de141bce173'
        )
        assert ret['type'] == 'dir'
        assert ret['contents'] == {
            'azcvsupdater_2.6.2.jar': {
                'priority': 4,
                'index': 0,
                'offset': 0,
                'progress': 0.0,
                'path': 'azcvsupdater_2.6.2.jar',
                'type': 'file',
                'size': 307949,
            }
        }

    @pytest_twisted.inlineCallbacks
    def test_download_torrent_from_url(self):
        filename = 'ubuntu-9.04-desktop-i386.iso.torrent'
        self.deluge_web.top_level.putChild(
            filename.encode(), File(common.get_test_data_file(filename))
        )
        url = 'http://localhost:%d/%s' % (self.webserver_listen_port, filename)
        res = yield self.deluge_web.web_api.download_torrent_from_url(url)
        assert res.endswith(filename)

    @pytest_twisted.inlineCallbacks
    def test_invalid_json(self):
        """
        If json_api._send_response does not return server.NOT_DONE_YET
        this error is thrown when json is invalid:
        exceptions.RuntimeError: Request.write called on a request after Request.finish was called.

        """
        agent = Agent(reactor)
        bad_body = b'{ method": "auth.login" }'
        d = yield agent.request(
            b'POST',
            b'http://127.0.0.1:%i/json' % self.webserver_listen_port,
            Headers(
                {
                    b'User-Agent': [b'Twisted Web Client Example'],
                    b'Content-Type': [b'application/json'],
                }
            ),
            FileBodyProducer(BytesIO(bad_body)),
        )
        yield d



#
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import errno
import logging
import os

from twisted.internet.utils import getProcessOutputAndValue
from twisted.python.procutils import which

import deluge.component as component
import deluge.configmanager
from deluge.common import windows_check
from deluge.core.rpcserver import export
from deluge.plugins.pluginbase import CorePluginBase

log = logging.getLogger(__name__)

DEFAULT_PREFS = {'extract_path': '', 'use_name_folder': True}

if windows_check():
    win_7z_exes = [
        '7z.exe',
        'C:\\Program Files\\7-Zip\\7z.exe',
        'C:\\Program Files (x86)\\7-Zip\\7z.exe',
    ]

    import winreg

    try:
        hkey = winreg.OpenKey(winreg.HKEY_CURRENT_USER, 'Software\\7-Zip')
    except OSError:
        pass
    else:
        win_7z_path = os.path.join(winreg.QueryValueEx(hkey, 'Path')[0], '7z.exe')
        winreg.CloseKey(hkey)
        win_7z_exes.insert(1, win_7z_path)

    switch_7z = 'x -y'
    # Future suport:
    # 7-zip cannot extract tar.* with single command.
    #    ".tar.gz", ".tgz",
    #    ".tar.bz2", ".tbz",
    #    ".tar.lzma", ".tlz",
    #    ".tar.xz", ".txz",
    exts_7z = ['.rar', '.zip', '.tar', '.7z', '.xz', '.lzma']
    for win_7z_exe in win_7z_exes:
        if which(win_7z_exe):
            EXTRACT_COMMANDS = dict.fromkeys(exts_7z, [win_7z_exe, switch_7z])
            break
else:
    required_cmds = ['unrar', 'unzip', 'tar', 'unxz', 'unlzma', '7zr', 'bunzip2']
    # Possible future suport:
    # gunzip: gz (cmd will delete original archive)
    # the following do not extract to dest dir
    # ".xz": ["xz", "-d --keep"],
    # ".lzma": ["xz", "-d --format=lzma --keep"],
    # ".bz2": ["bzip2", "-d --keep"],

    EXTRACT_COMMANDS = {
        '.rar': ['unrar', 'x -o+ -y'],
        '.tar': ['tar', '-xf'],
        '.zip': ['unzip', ''],
        '.tar.gz': ['tar', '-xzf'],
        '.tgz': ['tar', '-xzf'],
        '.tar.bz2': ['tar', '-xjf'],
        '.tbz': ['tar', '-xjf'],
        '.tar.lzma': ['tar', '--lzma -xf'],
        '.tlz': ['tar', '--lzma -xf'],
        '.tar.xz': ['tar', '--xz -xf'],
        '.txz': ['tar', '--xz -xf'],
        '.7z': ['7zr', 'x'],
    }
    # Test command exists and if not, remove.
    for command in required_cmds:
        if not which(command):
            for k, v in list(EXTRACT_COMMANDS.items()):
                if command in v[0]:
                    log.warning('%s not found, disabling support for %s', command, k)
                    del EXTRACT_COMMANDS[k]

if not EXTRACT_COMMANDS:
    raise Exception('No archive extracting programs found, plugin will be disabled')


class Core(CorePluginBase):
    def enable(self):
        self.config = deluge.configmanager.ConfigManager(
            'extractor.conf', DEFAULT_PREFS
        )
        if not self.config['extract_path']:
            self.config['extract_path'] = deluge.configmanager.ConfigManager(
                'core.conf'
            )['download_location']
        component.get('EventManager').register_event_handler(
            'TorrentFinishedEvent', self._on_torrent_finished
        )

    def disable(self):
        component.get('EventManager').deregister_event_handler(
            'TorrentFinishedEvent', self._on_torrent_finished
        )

    def update(self):
        pass

    def _on_torrent_finished(self, torrent_id):
        """
        This is called when a torrent finishes and checks if any files to extract.
        """
        tid = component.get('TorrentManager').torrents[torrent_id]
        tid_status = tid.get_status(['download_location', 'name'])

        files = tid.get_files()
        for f in files:
            file_root, file_ext = os.path.splitext(f['path'])
            file_ext_sec = os.path.splitext(file_root)[1]
            if file_ext_sec and file_ext_sec + file_ext in EXTRACT_COMMANDS:
                file_ext = file_ext_sec + file_ext
            elif file_ext not in EXTRACT_COMMANDS or file_ext_sec == '.tar':
                log.debug('Cannot extract file with unknown file type: %s', f['path'])
                continue
            elif file_ext == '.rar' and 'part' in file_ext_sec:
                part_num = file_ext_sec.split('part')[1]
                if part_num.isdigit() and int(part_num) != 1:
                    log.debug('Skipping remaining multi-part rar files: %s', f['path'])
                    continue

            cmd = EXTRACT_COMMANDS[file_ext]
            fpath = os.path.join(
                tid_status['download_location'], os.path.normpath(f['path'])
            )
            dest = os.path.normpath(self.config['extract_path'])
            if self.config['use_name_folder']:
                dest = os.path.join(dest, tid_status['name'])

            try:
                os.makedirs(dest)
            except OSError as ex:
                if not (ex.errno == errno.EEXIST and os.path.isdir(dest)):
                    log.error('Error creating destination folder: %s', ex)
                    break

            def on_extract(result, torrent_id, fpath):
                # Check command exit code.
                if not result[2]:
                    log.info('Extract successful: %s (%s)', fpath, torrent_id)
                else:
                    log.error(
                        'Extract failed: %s (%s) %s', fpath, torrent_id, result[1]
                    )

            # Run the command and add callback.
            log.debug(
                'Extracting %s from %s with %s %s to %s',
                fpath,
                torrent_id,
                cmd[0],
                cmd[1],
                dest,
            )
            d = getProcessOutputAndValue(
                cmd[0], cmd[1].split() + [str(fpath)], os.environ, str(dest)
            )
            d.addCallback(on_extract, torrent_id, fpath)

    @export
    def set_config(self, config):
        """Sets the config dictionary."""
        for key in config:
            self.config[key] = config[key]
        self.config.save()

    @export
    def get_config(self):
        """Returns the config dictionary."""
        return self.config.config



#
# Copyright (C) 2013 Bro <bro.development@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import deluge.component as component
import deluge.error
from deluge.common import get_localhost_auth
from deluge.conftest import BaseTestCase
from deluge.core import rpcserver
from deluge.core.authmanager import AuthManager
from deluge.core.rpcserver import DelugeRPCProtocol, RPCServer
from deluge.log import setup_logger

setup_logger('none')


class DelugeRPCProtocolTester(DelugeRPCProtocol):

    messages = []

    def transfer_message(self, data):
        self.messages.append(data)


class TestRPCServer(BaseTestCase):
    def set_up(self):
        self.rpcserver = RPCServer(listen=False)
        self.rpcserver.factory.protocol = DelugeRPCProtocolTester
        self.factory = self.rpcserver.factory
        self.session_id = '0'
        self.request_id = 11
        self.protocol = self.rpcserver.factory.protocol()
        self.protocol.factory = self.factory
        self.protocol.transport = self.protocol
        self.factory.session_protocols[self.session_id] = self.protocol
        self.factory.authorized_sessions[self.session_id] = None
        self.factory.interested_events[self.session_id] = ['TorrentFolderRenamedEvent']
        self.protocol.sessionno = self.session_id
        return component.start()

    def tear_down(self):
        def on_shutdown(result):
            del self.rpcserver

        return component.shutdown().addCallback(on_shutdown)

    def test_emit_event_for_session_id(self):
        torrent_id = '12'
        from deluge.event import TorrentFolderRenamedEvent

        data = [torrent_id, 'new name', 'old name']
        e = TorrentFolderRenamedEvent(*data)
        self.rpcserver.emit_event_for_session_id(self.session_id, e)
        msg = self.protocol.messages.pop()
        assert msg[0] == rpcserver.RPC_EVENT, str(msg)
        assert msg[1] == 'TorrentFolderRenamedEvent', str(msg)
        assert msg[2] == data, str(msg)

    def test_invalid_client_login(self):
        self.protocol.dispatch(self.request_id, 'daemon.login', [1], {})
        msg = self.protocol.messages.pop()
        assert msg[0] == rpcserver.RPC_ERROR
        assert msg[1] == self.request_id

    def test_valid_client_login(self):
        self.authmanager = AuthManager()
        auth = get_localhost_auth()
        self.protocol.dispatch(
            self.request_id, 'daemon.login', auth, {'client_version': 'Test'}
        )
        msg = self.protocol.messages.pop()
        assert msg[0] == rpcserver.RPC_RESPONSE, str(msg)
        assert msg[1] == self.request_id, str(msg)
        assert msg[2] == rpcserver.AUTH_LEVEL_ADMIN, str(msg)

    def test_client_login_error(self):
        # This test causes error log prints while running the test...
        self.protocol.transport = None  # This should cause AttributeError
        self.authmanager = AuthManager()
        auth = get_localhost_auth()
        self.protocol.dispatch(
            self.request_id, 'daemon.login', auth, {'client_version': 'Test'}
        )
        msg = self.protocol.messages.pop()
        assert msg[0] == rpcserver.RPC_ERROR
        assert msg[1] == self.request_id
        assert msg[2] == 'WrappedException'
        assert msg[3][1] == 'AttributeError'

    def test_client_invalid_method_call(self):
        self.authmanager = AuthManager()
        auth = get_localhost_auth()
        self.protocol.dispatch(self.request_id, 'invalid_function', auth, {})
        msg = self.protocol.messages.pop()
        assert msg[0] == rpcserver.RPC_ERROR
        assert msg[1] == self.request_id
        assert msg[2] == 'WrappedException'
        assert msg[3][1] == 'AttributeError'

    def test_daemon_info(self):
        self.protocol.dispatch(self.request_id, 'daemon.info', [], {})
        msg = self.protocol.messages.pop()
        assert msg[0] == rpcserver.RPC_RESPONSE, str(msg)
        assert msg[1] == self.request_id, str(msg)
        assert msg[2] == deluge.common.get_version(), str(msg)



#
# Copyright (C) 2007-2010 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging

import deluge.component as component

log = logging.getLogger(__name__)


class PluginBase(component.Component):

    update_interval = 1

    def __init__(self, name):
        super().__init__(name, self.update_interval)

    def enable(self):
        raise NotImplementedError('Need to define an enable method!')

    def disable(self):
        raise NotImplementedError('Need to define a disable method!')


class CorePluginBase(PluginBase):
    def __init__(self, plugin_name):
        super().__init__('CorePlugin.' + plugin_name)
        # Register RPC methods
        component.get('RPCServer').register_object(self, plugin_name.lower())
        log.debug('CorePlugin initialized..')

    def __del__(self):
        component.get('RPCServer').deregister_object(self)

    def enable(self):
        super().enable()

    def disable(self):
        super().disable()


class Gtk3PluginBase(PluginBase):
    def __init__(self, plugin_name):
        super().__init__('Gtk3Plugin.' + plugin_name)
        log.debug('Gtk3Plugin initialized..')

    def enable(self):
        super().enable()

    def disable(self):
        super().disable()


class WebPluginBase(PluginBase):

    scripts = []
    debug_scripts = []

    stylesheets = []
    debug_stylesheets = []

    def __init__(self, plugin_name):
        super().__init__('WebPlugin.' + plugin_name)

        # Register JSON rpc methods
        component.get('JSON').register_object(self, plugin_name.lower())
        log.debug('WebPlugin initialized..')

    def __del__(self):
        component.get('JSON').deregister_object(self)

    def enable(self):
        pass

    def disable(self):
        pass



#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#               2007-2009 Andrew Resch <andrewresch@gmail.com>
#               2009 Damien Churchill <damoxc@gmail.com>
#               2010 Pedro Algarvio <pedro@algarvio.me>
#               2017 Calum Lind <calumlind+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os.path

from pkg_resources import resource_filename


def get_resource(filename, subdir=False):
    folder = os.path.join('data', 'autoadd_options') if subdir else 'data'
    return resource_filename(__package__, os.path.join(folder, filename))



#!/usr/bin/env python
#
# -*- coding: utf-8 -*-
#
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from deluge.ui.client import sclient

sclient.set_core_uri()

print(sclient.get_enabled_plugins())

# enable plugin.
if 'label' not in sclient.get_enabled_plugins():
    sclient.enable_plugin('label')


# test labels.
print('#init labels')
try:
    sclient.label_remove('test')
except Exception:
    pass
sess_id = sclient.get_session_state()[0]

print('#add')
sclient.label_add('test')
print('#set')
sclient.label_set_torrent(id, 'test')

print(sclient.get_torrents_status({'label': 'test'}, 'name'))


print('#set options')
sclient.label_set_options('test', {'max_download_speed': 999}, True)
print(sclient.get_torrent_status(sess_id, ['max_download_speed']), '999')
sclient.label_set_options('test', {'max_download_speed': 9}, True)
print(sclient.get_torrent_status(sess_id, ['max_download_speed']), '9')
sclient.label_set_options('test', {'max_download_speed': 888}, False)
print(sclient.get_torrent_status(sess_id, ['max_download_speed']), '9 (888)')

print(sclient.get_torrent_status(sess_id, ['name', 'tracker_host', 'label']))



#
# Copyright (C) 2009-2010 Pedro Algarvio <pedro@algarvio.me>
#
# Basic plugin template created by:
# Copyright (C) 2008 Martijn Voncken <mvoncken@gmail.com>
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
# Copyright (C) 2009 Damien Churchill <damoxc@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import smtplib
from email.utils import formatdate

from twisted.internet import defer, threads

import deluge.configmanager
from deluge import component
from deluge.core.rpcserver import export
from deluge.event import known_events
from deluge.plugins.pluginbase import CorePluginBase

from .common import CustomNotifications

log = logging.getLogger(__name__)

DEFAULT_PREFS = {
    'smtp_enabled': False,
    'smtp_host': '',
    'smtp_port': 25,
    'smtp_user': '',
    'smtp_pass': '',
    'smtp_from': '',
    'smtp_tls': False,  # SSL or TLS
    'smtp_recipients': [],
    # Subscriptions
    'subscriptions': {'email': []},
}


class CoreNotifications(CustomNotifications):
    def __init__(self, plugin_name=None):
        CustomNotifications.__init__(self, plugin_name)

    def enable(self):
        CustomNotifications.enable(self)
        self.register_custom_email_notification(
            'TorrentFinishedEvent', self._on_torrent_finished_event
        )

    def disable(self):
        self.deregister_custom_email_notification('TorrentFinishedEvent')
        CustomNotifications.disable(self)

    def register_custom_email_notification(self, eventtype, handler):
        """This is used to register email notifications for custom event types.

        :param event: str, the event name
        :param handler: function, to be called when `:param:event` is emitted

        Your handler should return a tuple of (email_subject, email_contents).
        """
        self._register_custom_provider('email', eventtype, handler)

    def deregister_custom_email_notification(self, eventtype):
        self._deregister_custom_provider('email', eventtype)

    def handle_custom_email_notification(self, result, eventtype):
        if not self.config['smtp_enabled']:
            return defer.succeed('SMTP notification not enabled.')
        subject, message = result
        log.debug(
            'Spawning new thread to send email with subject: %s: %s', subject, message
        )
        # Spawn thread because we don't want Deluge to lock up while we send the
        # email.
        return threads.deferToThread(self._notify_email, subject, message)

    def get_handled_events(self):
        handled_events = []
        for evt in sorted(known_events):
            if known_events[evt].__module__.startswith('deluge.event'):
                if evt not in ('TorrentFinishedEvent',):
                    # Skip all un-handled built-in events
                    continue
            classdoc = known_events[evt].__doc__.strip()
            handled_events.append((evt, classdoc))
        log.debug('Handled Notification Events: %s', handled_events)
        return handled_events

    def _notify_email(self, subject='', message=''):
        log.debug('Email prepared')
        to_addrs = self.config['smtp_recipients']
        to_addrs_str = ', '.join(self.config['smtp_recipients'])
        headers_dict = {
            'smtp_from': self.config['smtp_from'],
            'subject': subject,
            'smtp_recipients': to_addrs_str,
            'date': formatdate(),
        }
        headers = (
            """\
From: %(smtp_from)s
To: %(smtp_recipients)s
Subject: %(subject)s
Date: %(date)s


"""
            % headers_dict
        )

        message = '\r\n'.join((headers + message).splitlines())

        try:
            server = smtplib.SMTP(
                self.config['smtp_host'], self.config['smtp_port'], timeout=60
            )
        except Exception as ex:
            err_msg = _('There was an error sending the notification email: %s') % ex
            log.error(err_msg)
            return ex

        security_enabled = self.config['smtp_tls']

        if security_enabled:
            server.ehlo()
            if 'starttls' not in server.esmtp_features:
                log.warning('TLS/SSL enabled but server does not support it')
            else:
                server.starttls()
                server.ehlo()

        if self.config['smtp_user'] and self.config['smtp_pass']:
            try:
                server.login(self.config['smtp_user'], self.config['smtp_pass'])
            except smtplib.SMTPHeloError as ex:
                err_msg = _('Server did not reply properly to HELO greeting: %s') % ex
                log.error(err_msg)
                return ex
            except smtplib.SMTPAuthenticationError as ex:
                err_msg = _('Server refused username/password combination: %s') % ex
                log.error(err_msg)
                return ex

        try:
            try:
                server.sendmail(self.config['smtp_from'], to_addrs, message.encode())
            except smtplib.SMTPException as ex:
                err_msg = (
                    _('There was an error sending the notification email: %s') % ex
                )
                log.error(err_msg)
                return ex
        finally:
            if security_enabled:
                # avoid false failure detection when the server closes
                # the SMTP connection with TLS enabled
                import socket

                try:
                    server.quit()
                except socket.sslerror:
                    pass
            else:
                server.quit()
        return _('Notification email sent.')

    def _on_torrent_finished_event(self, torrent_id):
        log.debug('Handler for TorrentFinishedEvent called for CORE')
        torrent = component.get('TorrentManager')[torrent_id]
        torrent_status = torrent.get_status(['name', 'num_files'])
        # Email
        subject = _('Finished Torrent "%(name)s"') % torrent_status
        message = (
            _(
                'This email is to inform you that Deluge has finished '
                'downloading "%(name)s", which includes %(num_files)i files.'
                '\nTo stop receiving these alerts, simply turn off email '
                "notification in Deluge's preferences.\n\n"
                'Thank you,\nDeluge.'
            )
            % torrent_status
        )
        return subject, message

        # d = defer.maybeDeferred(self.handle_custom_email_notification,
        #                        [subject, message],
        #                        'TorrentFinishedEvent')
        # d.addCallback(self._on_notify_sucess, 'email')
        # d.addErrback(self._on_notify_failure, 'email')
        # return d


class Core(CorePluginBase, CoreNotifications):
    def __init__(self, plugin_name):
        CorePluginBase.__init__(self, plugin_name)
        CoreNotifications.__init__(self)

    def enable(self):
        CoreNotifications.enable(self)
        self.config = deluge.configmanager.ConfigManager(
            'notifications-core.conf', DEFAULT_PREFS
        )
        log.debug('ENABLING CORE NOTIFICATIONS')

    def disable(self):
        log.debug('DISABLING CORE NOTIFICATIONS')
        CoreNotifications.disable(self)

    @export
    def set_config(self, config):
        """Sets the config dictionary."""
        for key in config:
            self.config[key] = config[key]
        self.config.save()

    @export
    def get_config(self):
        """Returns the config dictionary."""
        return self.config.config

    @export
    def get_handled_events(self):
        return CoreNotifications.get_handled_events(self)



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import json as json_lib
from io import BytesIO

import pytest_twisted
import twisted.web.client
from twisted.internet import reactor
from twisted.web.client import Agent, FileBodyProducer
from twisted.web.http_headers import Headers

from . import common
from .common import get_test_data_file
from .common_web import WebServerMockBase, WebServerTestBase

common.disable_new_release_check()


class TestWebServer(WebServerTestBase, WebServerMockBase):
    @pytest_twisted.inlineCallbacks
    def test_get_torrent_info(self):

        agent = Agent(reactor)

        self.mock_authentication_ignore(self.deluge_web.auth)

        # This torrent file contains an uncommon field 'filehash' which must be hex
        # encoded to allow dumping the torrent info to json. Otherwise it will fail with:
        # UnicodeDecodeError: 'utf8' codec can't decode byte 0xe5 in position 0: invalid continuation byte
        filename = get_test_data_file('filehash_field.torrent')
        input_file = (
            '{"params": ["%s"], "method": "web.get_torrent_info", "id": 22}'
            % filename.replace('\\', '\\\\')
        )
        headers = {
            b'User-Agent': ['Twisted Web Client Example'],
            b'Content-Type': ['application/json'],
        }
        url = 'http://127.0.0.1:%s/json' % self.webserver_listen_port

        d = yield agent.request(
            b'POST',
            url.encode('utf-8'),
            Headers(headers),
            FileBodyProducer(BytesIO(input_file.encode('utf-8'))),
        )
        body = yield twisted.web.client.readBody(d)

        try:
            json = json_lib.loads(body.decode())
        except Exception:
            print('aoeu')
        assert json['error'] is None
        assert 'torrent_filehash' == json['result']['name']



#
# Copyright (C) 2015 Calum Lind <calumlind@gmail.com>
# Copyright (C) 2010 Pedro Algarvio <ufs@ufsoft.org>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import warnings

from deluge.conftest import BaseTestCase
from deluge.log import setup_logger


class TestLog(BaseTestCase):
    def set_up(self):
        setup_logger(logging.DEBUG)

    def tear_down(self):
        setup_logger('none')

    def test_old_log_deprecation_warning(self):
        from deluge.log import LOG

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter('always')
            LOG.debug('foo')
            assert w[-1].category == DeprecationWarning

    # def test_twisted_error_log(self):
    #    from twisted.internet import defer
    #    import deluge.component as component
    #    from deluge.core.eventmanager import EventManager
    #    EventManager()
    #
    #    d = component.start()
    #
    #    @defer.inlineCallbacks
    #    def call(*args):
    #        yield component.pause(["EventManager"])
    #        yield component.start(["EventManager"])
    #
    #    d.addCallback(call)
    #    return d



#
# Copyright (C) 2009-2010 John Garland <johnnybg+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from .decompressers import BZipped2, GZipped, Zipped
from .readers import EmuleReader, PeerGuardianReader, SafePeerReader

COMPRESSION_TYPES = {b'PK': 'Zip', b'\x1f\x8b': 'GZip', b'BZ': 'BZip2'}

DECOMPRESSERS = {'Zip': Zipped, 'GZip': GZipped, 'BZip2': BZipped2}

READERS = {
    'Emule': EmuleReader,
    'SafePeer': SafePeerReader,
    'PeerGuardian': PeerGuardianReader,
}


class UnknownFormatError(Exception):
    pass


def detect_compression(filename):
    with open(filename, 'rb') as _file:
        magic_number = _file.read(2)
    return COMPRESSION_TYPES.get(magic_number, '')


def detect_format(filename, compression=''):
    file_format = ''
    for reader in READERS:
        if create_reader(reader, compression)(filename).is_valid():
            file_format = reader
            break
    return file_format


def create_reader(file_format, compression=''):
    reader = READERS.get(file_format)
    if reader and compression:
        decompressor = DECOMPRESSERS.get(compression)
        if decompressor:
            reader = decompressor(reader)
    return reader



#
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import logging
import os
from socket import gaierror, getaddrinfo
from urllib.parse import urlparse

from gi.repository import Gtk
from twisted.internet import defer, reactor

import deluge.component as component
from deluge.common import resource_filename, windows_check
from deluge.configmanager import ConfigManager, get_config_dir
from deluge.error import AuthenticationRequired, BadLoginError, IncompatibleClient
from deluge.ui.client import Client, client
from deluge.ui.hostlist import DEFAULT_PORT, LOCALHOST, HostList

from .common import get_clipboard_text
from .dialogs import AuthenticationDialog, ErrorDialog

log = logging.getLogger(__name__)

HOSTLIST_COL_ID = 0
HOSTLIST_COL_HOST = 1
HOSTLIST_COL_PORT = 2
HOSTLIST_COL_USER = 3
HOSTLIST_COL_PASS = 4
HOSTLIST_COL_STATUS = 5
HOSTLIST_COL_VERSION = 6
HOSTLIST_COL_STATUS_I18N = 7

HOSTLIST_ICONS = {
    'offline': 'action-unavailable-symbolic',
    'online': 'network-server-symbolic',
    'connected': 'network-transmit-receive-symbolic',
}
STATUS_I18N = {
    'offline': _('Offline'),
    'online': _('Online'),
    'connected': _('Connected'),
}


def cell_render_host(column, cell, model, row, data):
    host, port, username = model.get(row, *data)
    text = host + ':' + str(port)
    if username:
        text = username + '@' + text
    cell.set_property('text', text)


def cell_render_status_icon(column, cell, model, row, data):
    status = model[row][data]
    status = status if status else 'offline'
    icon_name = HOSTLIST_ICONS.get(status, None)
    cell.set_property('icon-name', icon_name)


class ConnectionManager(component.Component):
    def __init__(self):
        component.Component.__init__(self, 'ConnectionManager')
        self.gtkui_config = ConfigManager('gtk3ui.conf')
        self.hostlist = HostList()
        self.running = False

    # Component overrides
    def start(self):
        pass

    def stop(self):
        # Close this dialog when we are shutting down
        if self.running:
            self.connection_manager.response(Gtk.ResponseType.CLOSE)

    def shutdown(self):
        pass

    # Public methods
    def show(self):
        """Show the ConnectionManager dialog."""
        self.builder = Gtk.Builder()
        self.builder.add_from_file(
            resource_filename(
                __package__, os.path.join('glade', 'connection_manager.ui')
            )
        )
        self.connection_manager = self.builder.get_object('connection_manager')
        self.connection_manager.set_transient_for(component.get('MainWindow').window)

        # Setup the hostlist liststore and treeview
        self.treeview = self.builder.get_object('treeview_hostlist')
        self.treeview.set_tooltip_column(HOSTLIST_COL_STATUS_I18N)
        self.liststore = self.builder.get_object('liststore_hostlist')

        render = Gtk.CellRendererPixbuf()
        column = Gtk.TreeViewColumn(_('Status'), render)
        column.set_cell_data_func(render, cell_render_status_icon, HOSTLIST_COL_STATUS)
        self.treeview.append_column(column)

        render = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_('Host'), render, text=HOSTLIST_COL_HOST)
        host_data = (HOSTLIST_COL_HOST, HOSTLIST_COL_PORT, HOSTLIST_COL_USER)
        column.set_cell_data_func(render, cell_render_host, host_data)
        column.set_expand(True)
        self.treeview.append_column(column)

        column = Gtk.TreeViewColumn(
            _('Version'), Gtk.CellRendererText(), text=HOSTLIST_COL_VERSION
        )
        self.treeview.append_column(column)

        # Load any saved host entries
        self._load_liststore()
        # Set widgets to values from gtkui config.
        self._load_widget_config()
        self._update_widget_buttons()

        # Connect the signals to the handlers
        self.builder.connect_signals(self)
        self.treeview.get_selection().connect(
            'changed', self.on_hostlist_selection_changed
        )

        # Set running True before update status call.
        self.running = True

        if windows_check():
            # Call to simulate() required to workaround showing daemon status (see #2813)
            reactor.simulate()
        self._update_host_status()

        # Trigger the on_selection_changed code and select the first host if possible
        self.treeview.get_selection().unselect_all()
        if len(self.liststore):
            self.treeview.get_selection().select_path(0)

        # Run the dialog
        self.connection_manager.run()

        # Dialog closed so cleanup.
        self.running = False
        self.connection_manager.destroy()
        del self.builder
        del self.connection_manager
        del self.liststore
        del self.treeview

    def _load_liststore(self):
        """Load saved host entries"""
        for host_entry in self.hostlist.get_hosts_info():
            host_id, host, port, username = host_entry
            self.liststore.append([host_id, host, port, username, '', '', '', ''])

    def _load_widget_config(self):
        """Set the widgets to show the correct options from the config."""
        self.builder.get_object('chk_autoconnect').set_active(
            self.gtkui_config['autoconnect']
        )
        self.builder.get_object('chk_autostart').set_active(
            self.gtkui_config['autostart_localhost']
        )
        self.builder.get_object('chk_donotshow').set_active(
            not self.gtkui_config['show_connection_manager_on_start']
        )

    def _update_host_status(self):
        """Updates the host status"""
        if not self.running:
            # Callback likely fired after the window closed.
            return

        def on_host_status(status_info, row):
            if self.running and row:
                status = status_info[1].lower()
                row[HOSTLIST_COL_STATUS] = status
                row[HOSTLIST_COL_STATUS_I18N] = STATUS_I18N[status]
                row[HOSTLIST_COL_VERSION] = status_info[2]
                self._update_widget_buttons()

        deferreds = []
        for row in self.liststore:
            host_id = row[HOSTLIST_COL_ID]
            d = self.hostlist.get_host_status(host_id)
            try:
                d.addCallback(on_host_status, row)
            except AttributeError:
                on_host_status(d, row)
            else:
                deferreds.append(d)
        defer.DeferredList(deferreds)

    def _update_widget_buttons(self):
        """Updates the dialog button states."""
        self.builder.get_object('button_refresh').set_sensitive(len(self.liststore))
        self.builder.get_object('button_startdaemon').set_sensitive(False)
        self.builder.get_object('button_connect').set_sensitive(False)
        self.builder.get_object('button_connect').set_label(_('C_onnect'))
        self.builder.get_object('button_edithost').set_sensitive(False)
        self.builder.get_object('button_removehost').set_sensitive(False)
        self.builder.get_object('button_startdaemon').set_sensitive(False)
        self.builder.get_object('image_startdaemon').set_from_icon_name(
            'system-run-symbolic', Gtk.IconSize.BUTTON
        )
        self.builder.get_object('label_startdaemon').set_text_with_mnemonic(
            _('_Start Daemon')
        )

        model, row = self.treeview.get_selection().get_selected()
        if row:
            self.builder.get_object('button_edithost').set_sensitive(True)
            self.builder.get_object('button_removehost').set_sensitive(True)
        else:
            return

        # Get selected host info.
        __, host, port, __, __, status, __, __ = model[row]

        try:
            getaddrinfo(host, None)
        except gaierror as ex:
            log.error(
                'Error resolving host %s to ip: %s', row[HOSTLIST_COL_HOST], ex.args[1]
            )
            self.builder.get_object('button_connect').set_sensitive(False)
            return

        log.debug('Host Status: %s, %s', host, status)

        # Check to see if the host is online
        if status == 'connected' or status == 'online':
            self.builder.get_object('button_connect').set_sensitive(True)
            self.builder.get_object('image_startdaemon').set_from_icon_name(
                'process-stop-symbolic', Gtk.IconSize.MENU
            )
            self.builder.get_object('label_startdaemon').set_text_with_mnemonic(
                _('_Stop Daemon')
            )
            self.builder.get_object('button_startdaemon').set_sensitive(False)
            if status == 'connected':
                # Display a disconnect button if we're connected to this host
                self.builder.get_object('button_connect').set_label(_('_Disconnect'))
                self.builder.get_object('button_removehost').set_sensitive(False)
                # Currently can only stop daemon when connected to it
                self.builder.get_object('button_startdaemon').set_sensitive(True)
        elif host in LOCALHOST:
            # If localhost we can start the dameon.
            self.builder.get_object('button_startdaemon').set_sensitive(True)

    def start_daemon(self, port, config):
        """Attempts to start local daemon process and will show an ErrorDialog if not.

        Args:
            port (int): Port for the daemon to listen on.
            config (str): Config path to pass to daemon.

        Returns:
            bool: True is successfully started the daemon, False otherwise.

        """
        if client.start_daemon(port, config):
            log.debug('Localhost daemon started')
            reactor.callLater(1, self._update_host_status)
            return True
        else:
            ErrorDialog(
                _('Unable to start daemon!'),
                _('Check deluged package is installed and logs for further details'),
            ).run()
            return False

    # Signal handlers
    def _connect(self, host_id, username=None, password=None, try_counter=0):
        def do_connect(result, username=None, password=None, *args):
            log.debug('Attempting to connect to daemon...')
            for host_entry in self.hostlist.config['hosts']:
                if host_entry[0] == host_id:
                    __, host, port, host_user, host_pass = host_entry

            username = username if username else host_user
            password = password if password else host_pass

            d = client.connect(host, port, username, password)
            d.addCallback(self._on_connect, host_id)
            d.addErrback(self._on_connect_fail, host_id, try_counter)
            return d

        if client.connected():
            return client.disconnect().addCallback(do_connect, username, password)
        else:
            return do_connect(None, username, password)

    def _on_connect(self, daemon_info, host_id):
        log.debug('Connected to daemon: %s', host_id)
        if self.gtkui_config['autoconnect']:
            self.gtkui_config['autoconnect_host_id'] = host_id
        if self.running:
            # When connected to a client, and then trying to connect to another,
            # this component will be stopped(while the connect deferred is
            # running), so, self.connection_manager will be deleted.
            # If that's not the case, close the dialog.
            self.connection_manager.response(Gtk.ResponseType.OK)
        component.start()

    def _on_connect_fail(self, reason, host_id, try_counter):
        log.debug('Failed to connect: %s', reason.value)

        if reason.check(AuthenticationRequired, BadLoginError):
            log.debug('PasswordRequired exception')
            dialog = AuthenticationDialog(reason.value.message, reason.value.username)

            def dialog_finished(response_id):
                if response_id == Gtk.ResponseType.OK:
                    self._connect(host_id, dialog.get_username(), dialog.get_password())

            return dialog.run().addCallback(dialog_finished)

        elif reason.check(IncompatibleClient):
            return ErrorDialog(_('Incompatible Client'), reason.value.message).run()

        if try_counter:
            log.info('Retrying connection.. Retries left: %s', try_counter)
            return reactor.callLater(
                0.5, self._connect, host_id, try_counter=try_counter - 1
            )

        msg = str(reason.value)
        if not self.gtkui_config['autostart_localhost']:
            msg += '\n' + _(
                'Auto-starting the daemon locally is not enabled. '
                'See "Options" on the "Connection Manager".'
            )
        ErrorDialog(_('Failed To Connect'), msg).run()

    def on_button_connect_clicked(self, widget=None):
        """Button handler for connect to or disconnect from daemon."""
        model, row = self.treeview.get_selection().get_selected()
        if not row:
            return

        host_id, host, port, __, __, status, __, __ = model[row]
        # If status is connected then connect button disconnects instead.
        if status == 'connected':

            def on_disconnect(reason):
                self._update_host_status()

            return client.disconnect().addCallback(on_disconnect)

        try_counter = 0
        auto_start = self.builder.get_object('chk_autostart').get_active()
        if auto_start and host in LOCALHOST and status == 'offline':
            # Start the local daemon and then connect with retries set.
            if self.start_daemon(port, get_config_dir()):
                try_counter = 6
            else:
                # Don't attempt to connect to offline daemon.
                return

        self._connect(host_id, try_counter=try_counter)

    def on_button_close_clicked(self, widget):
        self.connection_manager.response(Gtk.ResponseType.CLOSE)

    def _run_addhost_dialog(self, edit_host_info=None):
        """Create and runs the add host dialog.

        Supplying edit_host_info changes the dialog to an edit dialog.

        Args:
            edit_host_info (list): A list of (host, port, user, pass) to edit.

        Returns:
            list: The new host info values (host, port, user, pass).

        """
        self.builder.add_from_file(
            resource_filename(
                __package__, os.path.join('glade', 'connection_manager.addhost.ui')
            )
        )
        dialog = self.builder.get_object('addhost_dialog')
        dialog.set_transient_for(self.connection_manager)
        hostname_entry = self.builder.get_object('entry_hostname')
        port_spinbutton = self.builder.get_object('spinbutton_port')
        username_entry = self.builder.get_object('entry_username')
        password_entry = self.builder.get_object('entry_password')

        if edit_host_info:
            dialog.set_title(_('Edit Host'))
            hostname_entry.set_text(edit_host_info[0])
            port_spinbutton.set_value(edit_host_info[1])
            username_entry.set_text(edit_host_info[2])
            password_entry.set_text(edit_host_info[3])

        response = dialog.run()
        new_host_info = []
        if response:
            new_host_info.append(hostname_entry.get_text())
            new_host_info.append(port_spinbutton.get_value_as_int())
            new_host_info.append(username_entry.get_text())
            new_host_info.append(password_entry.get_text())

        dialog.destroy()
        return new_host_info

    def on_button_addhost_clicked(self, widget):
        log.debug('on_button_addhost_clicked')
        host_info = self._run_addhost_dialog()
        if host_info:
            hostname, port, username, password = host_info
            try:
                host_id = self.hostlist.add_host(hostname, port, username, password)
            except ValueError as ex:
                ErrorDialog(_('Error Adding Host'), ex).run()
            else:
                status = 'offline'
                version = ''
                self.liststore.append(
                    [
                        host_id,
                        hostname,
                        port,
                        username,
                        password,
                        status,
                        version,
                        STATUS_I18N[status],
                    ]
                )
                self._update_host_status()

    def on_button_edithost_clicked(self, widget=None):
        log.debug('on_button_edithost_clicked')
        model, row = self.treeview.get_selection().get_selected()
        status = model[row][HOSTLIST_COL_STATUS]
        host_id = model[row][HOSTLIST_COL_ID]
        host_info = [
            self.liststore[row][HOSTLIST_COL_HOST],
            self.liststore[row][HOSTLIST_COL_PORT],
            self.liststore[row][HOSTLIST_COL_USER],
            self.liststore[row][HOSTLIST_COL_PASS],
        ]

        new_host_info = self._run_addhost_dialog(edit_host_info=host_info)
        if new_host_info:
            hostname, port, username, password = new_host_info
            try:
                self.hostlist.update_host(host_id, hostname, port, username, password)
            except ValueError as ex:
                ErrorDialog(_('Error Updating Host'), ex).run()
            else:
                self.liststore[row] = (
                    host_id,
                    hostname,
                    port,
                    username,
                    password,
                    '',
                    '',
                    '',
                )
                self._update_host_status()

            if status == 'connected':

                def on_disconnect(reason):
                    self._update_host_status()

                client.disconnect().addCallback(on_disconnect)

    def on_button_removehost_clicked(self, widget):
        log.debug('on_button_removehost_clicked')
        # Get the selected rows
        model, row = self.treeview.get_selection().get_selected()
        self.hostlist.remove_host(model[row][HOSTLIST_COL_ID])
        self.liststore.remove(row)
        # Update the hostlist
        self._update_host_status()

    def on_button_startdaemon_clicked(self, widget):
        log.debug('on_button_startdaemon_clicked')
        if not self.liststore.iter_n_children(None):
            # There is nothing in the list, so lets create a localhost entry
            try:
                self.hostlist.add_default_host()
            except ValueError as ex:
                log.error('Error adding default host: %s', ex)
            else:
                self.start_daemon(DEFAULT_PORT, get_config_dir())
            finally:
                return

        paths = self.treeview.get_selection().get_selected_rows()[1]
        if len(paths):
            __, host, port, user, password, status, __, __ = self.liststore[paths[0]]
        else:
            return

        if host not in LOCALHOST:
            return

        def on_daemon_status_change(result):
            """Daemon start/stop callback"""
            reactor.callLater(0.7, self._update_host_status)

        if status in ('online', 'connected'):
            # Button will stop the daemon if status is online or connected.
            def on_connect(d, c):
                """Client callback to call daemon shutdown"""
                c.daemon.shutdown().addCallback(on_daemon_status_change)

            if client.connected() and (host, port, user) == client.connection_info():
                client.daemon.shutdown().addCallback(on_daemon_status_change)
            elif user and password:
                c = Client()
                c.connect(host, port, user, password).addCallback(on_connect, c)
        else:
            # Otherwise button will start the daemon.
            self.start_daemon(port, get_config_dir())

    def on_button_refresh_clicked(self, widget):
        self._update_host_status()

    def on_hostlist_row_activated(self, tree, path, view_column):
        self.on_button_connect_clicked()

    def on_hostlist_selection_changed(self, treeselection):
        self._update_widget_buttons()

    def on_chk_toggled(self, widget):
        self.gtkui_config['autoconnect'] = self.builder.get_object(
            'chk_autoconnect'
        ).get_active()
        self.gtkui_config['autostart_localhost'] = self.builder.get_object(
            'chk_autostart'
        ).get_active()
        self.gtkui_config[
            'show_connection_manager_on_start'
        ] = not self.builder.get_object('chk_donotshow').get_active()

    def on_entry_host_paste_clipboard(self, widget):
        text = get_clipboard_text()
        log.debug('on_entry_proxy_host_paste-clipboard: got paste: %s', text)
        text = text if '//' in text else '//' + text
        parsed = urlparse(text)
        if parsed.hostname:
            widget.set_text(parsed.hostname)
            widget.emit_stop_by_name('paste-clipboard')
        if parsed.port:
            self.builder.get_object('spinbutton_port').set_value(parsed.port)
        if parsed.username:
            self.builder.get_object('entry_username').set_text(parsed.username)
        if parsed.password:
            self.builder.get_object('entry_password').set_text(parsed.password)



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#
import pytest
import pytest_twisted
import twisted.python.failure
from twisted.internet import defer, reactor, task
from twisted.internet.defer import maybeDeferred

from deluge.decorators import maybe_coroutine


@defer.inlineCallbacks
def inline_func():
    result = yield task.deferLater(reactor, 0, lambda: 'function_result')
    return result


@defer.inlineCallbacks
def inline_error():
    raise Exception('function_error')
    yield


@maybe_coroutine
async def coro_func():
    result = await task.deferLater(reactor, 0, lambda: 'function_result')
    return result


@maybe_coroutine
async def coro_error():
    raise Exception('function_error')


@defer.inlineCallbacks
def coro_func_from_inline():
    result = yield coro_func()
    return result


@defer.inlineCallbacks
def coro_error_from_inline():
    result = yield coro_error()
    return result


@maybe_coroutine
async def coro_func_from_coro():
    return await coro_func()


@maybe_coroutine
async def coro_error_from_coro():
    return await coro_error()


@maybe_coroutine
async def inline_func_from_coro():
    return await inline_func()


@maybe_coroutine
async def inline_error_from_coro():
    return await inline_error()


@pytest_twisted.inlineCallbacks
def test_standard_twisted():
    """Sanity check that twisted tests work how we expect.

    Not really testing deluge code at all.
    """
    result = yield inline_func()
    assert result == 'function_result'

    with pytest.raises(Exception, match='function_error'):
        yield inline_error()


@pytest.mark.parametrize(
    'function',
    [
        inline_func,
        coro_func,
        coro_func_from_coro,
        coro_func_from_inline,
        inline_func_from_coro,
    ],
)
@pytest_twisted.inlineCallbacks
def test_from_inline(function):
    """Test our coroutines wrapped with maybe_coroutine as if they returned plain twisted deferreds."""
    result = yield function()
    assert result == 'function_result'

    def cb(result):
        assert result == 'function_result'

    d = function()
    d.addCallback(cb)
    yield d


@pytest.mark.parametrize(
    'function',
    [
        inline_error,
        coro_error,
        coro_error_from_coro,
        coro_error_from_inline,
        inline_error_from_coro,
    ],
)
@pytest_twisted.inlineCallbacks
def test_error_from_inline(function):
    """Test our coroutines wrapped with maybe_coroutine as if they returned plain twisted deferreds that raise."""
    with pytest.raises(Exception, match='function_error'):
        yield function()

    def eb(result):
        assert isinstance(result, twisted.python.failure.Failure)
        assert result.getErrorMessage() == 'function_error'

    d = function()
    d.addErrback(eb)
    yield d


@pytest.mark.parametrize(
    'function',
    [
        inline_func,
        coro_func,
        coro_func_from_coro,
        coro_func_from_inline,
        inline_func_from_coro,
    ],
)
@pytest_twisted.ensureDeferred
async def test_from_coro(function):
    """Test our coroutines wrapped with maybe_coroutine work from another coroutine."""
    result = await function()
    assert result == 'function_result'


@pytest.mark.parametrize(
    'function',
    [
        inline_error,
        coro_error,
        coro_error_from_coro,
        coro_error_from_inline,
        inline_error_from_coro,
    ],
)
@pytest_twisted.ensureDeferred
async def test_error_from_coro(function):
    """Test our coroutines wrapped with maybe_coroutine work from another coroutine with errors."""
    with pytest.raises(Exception, match='function_error'):
        await function()


@pytest_twisted.ensureDeferred
async def test_tracebacks_preserved():
    with pytest.raises(Exception) as exc:
        await coro_error_from_coro()
    traceback_lines = [
        'await coro_error_from_coro()',
        'return await coro_error()',
        "raise Exception('function_error')",
    ]
    # If each coroutine got wrapped with ensureDeferred, the traceback will be mangled
    # verify the coroutines passed through by checking the traceback.
    for expected, actual in zip(traceback_lines, exc.traceback):
        assert expected in str(actual)


@pytest_twisted.ensureDeferred
async def test_maybe_deferred_coroutine():
    result = await maybeDeferred(coro_func)
    assert result == 'function_result'


@pytest_twisted.ensureDeferred
async def test_callback_before_await():
    def cb(res):
        assert res == 'function_result'
        return res

    d = coro_func()
    d.addCallback(cb)
    result = await d
    assert result == 'function_result'


@pytest_twisted.ensureDeferred
async def test_callback_after_await():
    """If it has already been used as a coroutine, can't be retroactively turned into a Deferred.
    This limitation could be fixed, but the extra complication doesn't feel worth it.
    """

    def cb(res):
        pass

    d = coro_func()
    await d
    with pytest.raises(
        Exception, match='Cannot add callbacks to an already awaited coroutine'
    ):
        d.addCallback(cb)



#
# Copyright (C) 2007-2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from deluge.plugins.init import PluginInitBase


class CorePlugin(PluginInitBase):
    def __init__(self, plugin_name):
        from .core import Core as _pluginCls

        self._plugin_cls = _pluginCls
        super().__init__(plugin_name)


class GtkUIPlugin(PluginInitBase):
    def __init__(self, plugin_name):
        from .gtkui import GtkUI as _pluginCls

        self._plugin_cls = _pluginCls
        super().__init__(plugin_name)


class WebUIPlugin(PluginInitBase):
    def __init__(self, plugin_name):
        from .webui import WebUI as _pluginCls

        self._plugin_cls = _pluginCls
        super().__init__(plugin_name)



#
# Copyright (C) 2016 bendikro <bro.devel+deluge@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

from deluge.ui.common import TorrentInfo

from . import common


class TestUICommon:
    def test_hash_optional_single_file(self):
        """Ensure single file with `ed2k` and `sha1` keys are not in filetree output."""
        filename = common.get_test_data_file('test.torrent')
        files_tree = {'azcvsupdater_2.6.2.jar': (0, 307949, True)}
        ti = TorrentInfo(filename, filetree=1)
        assert ti.files_tree == files_tree

        files_tree2 = {
            'contents': {
                'azcvsupdater_2.6.2.jar': {
                    'type': 'file',
                    'index': 0,
                    'length': 307949,
                    'download': True,
                }
            }
        }
        ti = TorrentInfo(filename, filetree=2)
        assert ti.files_tree == files_tree2

    def test_hash_optional_multi_file(self):
        """Ensure multi-file with `filehash` and `ed2k` are keys not in filetree output."""
        filename = common.get_test_data_file('filehash_field.torrent')
        files_tree = {
            'torrent_filehash': {
                'tull.txt': (0, 54, True),
                '還在一個人無聊嗎~還不趕緊上來聊天美.txt': (1, 54, True),
            }
        }
        ti = TorrentInfo(filename, filetree=1)
        assert ti.files_tree == files_tree

        filestree2 = {
            'contents': {
                'torrent_filehash': {
                    'type': 'dir',
                    'contents': {
                        'tull.txt': {
                            'type': 'file',
                            'path': 'torrent_filehash/tull.txt',
                            'length': 54,
                            'index': 0,
                            'download': True,
                        },
                        '還在一個人無聊嗎~還不趕緊上來聊天美.txt': {
                            'type': 'file',
                            'path': 'torrent_filehash/還在一個人無聊嗎~還不趕緊上來聊天美.txt',
                            'length': 54,
                            'index': 1,
                            'download': True,
                        },
                    },
                    'length': 108,
                    'download': True,
                }
            },
            'type': 'dir',
        }
        ti = TorrentInfo(filename, filetree=2)
        assert ti.files_tree == filestree2

    def test_hash_optional_md5sum(self):
        # Ensure `md5sum` key is not included in filetree output
        filename = common.get_test_data_file('md5sum.torrent')
        files_tree = {'test': {'lol': (0, 4, True), 'rofl': (1, 5, True)}}
        ti = TorrentInfo(filename, filetree=1)
        assert ti.files_tree == files_tree
        ti = TorrentInfo(filename, filetree=2)
        files_tree2 = {
            'contents': {
                'test': {
                    'type': 'dir',
                    'contents': {
                        'lol': {
                            'type': 'file',
                            'path': 'test/lol',
                            'index': 0,
                            'length': 4,
                            'download': True,
                        },
                        'rofl': {
                            'type': 'file',
                            'path': 'test/rofl',
                            'index': 1,
                            'length': 5,
                            'download': True,
                        },
                    },
                    'length': 9,
                    'download': True,
                }
            },
            'type': 'dir',
        }
        assert ti.files_tree == files_tree2

    def test_utf8_encoded_paths(self):
        filename = common.get_test_data_file('test.torrent')
        ti = TorrentInfo(filename)
        assert 'azcvsupdater_2.6.2.jar' in ti.files_tree

    def test_utf8_encoded_paths2(self):
        filename = common.get_test_data_file('unicode_filenames.torrent')
        filepath1 = '\u30c6\u30af\u30b9\u30fb\u30c6\u30af\u30b5\u30f3.mkv'
        filepath2 = (
            '\u041c\u0438\u0445\u0430\u0438\u043b \u0413\u043e'
            '\u0440\u0431\u0430\u0447\u0451\u0432.mkv'
        )
        filepath3 = "Alisher ibn G'iyosiddin Navoiy.mkv"
        filepath4 = 'Ascii title.mkv'
        filepath5 = '\u09b8\u09c1\u0995\u09c1\u09ae\u09be\u09b0 \u09b0\u09be\u09df.mkv'

        ti = TorrentInfo(filename)
        files_tree = ti.files_tree['unicode_filenames']
        assert filepath1 in files_tree
        assert filepath2 in files_tree
        assert filepath3 in files_tree
        assert filepath4 in files_tree
        assert filepath5 in files_tree

        result_files = [
            {
                'download': True,
                'path': 'unicode_filenames/' + filepath3,
                'size': 126158658,
            },
            {
                'download': True,
                'path': 'unicode_filenames/' + filepath4,
                'size': 189321363,
            },
            {
                'download': True,
                'path': 'unicode_filenames/' + filepath2,
                'size': 106649699,
            },
            {
                'download': True,
                'path': 'unicode_filenames/' + filepath5,
                'size': 21590269,
            },
            {'download': True, 'path': 'unicode_filenames/' + filepath1, 'size': 1771},
        ]

        assert len(ti.files) == len(result_files)



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import deluge.error


class TestError:
    def test_deluge_error(self):
        msg = 'Some message'
        e = deluge.error.DelugeError(msg)
        assert str(e) == msg
        from twisted.internet.defer import DebugInfo

        del DebugInfo.__del__  # Hides all errors
        assert e._args == (msg,)
        assert e._kwargs == {}

    def test_incompatible_client(self):
        version = '1.3.6'
        e = deluge.error.IncompatibleClient(version)
        assert (
            str(e) == 'Your deluge client is not compatible with the daemon. '
            'Please upgrade your client to %s' % version
        )

    def test_not_authorized_error(self):
        current_level = 5
        required_level = 10
        e = deluge.error.NotAuthorizedError(current_level, required_level)
        assert str(e) == 'Auth level too low: %d < %d' % (current_level, required_level)

    def test_bad_login_error(self):
        message = 'Login failed'
        username = 'deluge'
        e = deluge.error.BadLoginError(message, username)
        assert str(e) == message



#
# Copyright (C) 2009 Andrew Resch <andrewresch@gmail.com>
#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
from hashlib import sha1 as sha

from deluge.bencode import bencode
from deluge.common import get_path_size, utf8_encode_structure


class InvalidPath(Exception):
    """Raised when an invalid path is supplied."""

    pass


class InvalidPieceSize(Exception):
    """Raised when an invalid piece size is set.

    Note:
        Piece sizes must be multiples of 16KiB.
    """

    pass


class TorrentMetadata:
    """This class is used to create .torrent files.

    Examples:

        >>> t = TorrentMetadata()
        >>> t.data_path = '/tmp/torrent'
        >>> t.comment = 'My Test Torrent'
        >>> t.trackers = [['http://tracker.openbittorent.com']]
        >>> t.save('/tmp/test.torrent')

    """

    def __init__(self):
        self.__data_path = None
        self.__piece_size = 0
        self.__comment = ''
        self.__private = False
        self.__trackers = []
        self.__webseeds = []
        self.__pad_files = False

    def save(self, torrent_path, progress=None):
        """Creates and saves the torrent file to `path`.

        Args:
            torrent_path (str): Location to save the torrent file.
            progress(func, optional): The function to be called when a piece is hashed. The
                provided function should be in the format `func(num_completed, num_pieces)`.

        Raises:
            InvalidPath: If the data_path has not been set.

        """
        if not self.data_path:
            raise InvalidPath('Need to set a data_path!')

        torrent = {'info': {}}

        if self.comment:
            torrent['comment'] = self.comment

        if self.private:
            torrent['info']['private'] = True

        if self.trackers:
            torrent['announce'] = self.trackers[0][0]
            torrent['announce-list'] = self.trackers
        else:
            torrent['announce'] = ''

        if self.webseeds:
            httpseeds = []
            webseeds = []
            for w in self.webseeds:
                if w.endswith('.php'):
                    httpseeds.append(w)
                else:
                    webseeds.append(w)

            if httpseeds:
                torrent['httpseeds'] = httpseeds
            if webseeds:
                torrent['url-list'] = webseeds

        datasize = get_path_size(self.data_path)

        if self.piece_size:
            piece_size = self.piece_size * 1024
        else:
            # We need to calculate a piece size
            piece_size = 16384
            while (datasize // piece_size) > 1024 and piece_size < (8192 * 1024):
                piece_size *= 2

        # Calculate the number of pieces we will require for the data
        num_pieces = datasize // piece_size
        if datasize % piece_size:
            num_pieces += 1

        torrent['info']['piece length'] = piece_size
        torrent['info']['name'] = os.path.split(self.data_path)[1]

        # Create the info
        if os.path.isdir(self.data_path):
            files = []
            padding_count = 0
            # Collect a list of file paths and add padding files if necessary
            for (dirpath, dirnames, filenames) in os.walk(self.data_path):
                for index, filename in enumerate(filenames):
                    size = get_path_size(
                        os.path.join(self.data_path, dirpath, filename)
                    )
                    p = dirpath[len(self.data_path) :]
                    p = p.lstrip('/')
                    p = p.split('/')
                    if p[0]:
                        p += [filename]
                    else:
                        p = [filename]
                    files.append((size, p))
                    # Add a padding file if necessary
                    if self.pad_files and (index + 1) < len(filenames):
                        left = size % piece_size
                        if left:
                            p = list(p)
                            p[-1] = '_____padding_file_' + str(padding_count)
                            files.append((piece_size - left, p))
                            padding_count += 1

            # Run the progress function with 0 completed pieces
            if progress:
                progress(0, num_pieces)

            fs = []
            pieces = []
            # Create the piece hashes
            buf = b''
            for size, path in files:
                path = [s.encode('UTF-8') for s in path]
                fs.append({b'length': size, b'path': path})
                if path[-1].startswith(b'_____padding_file_'):
                    buf += b'\0' * size
                    pieces.append(sha(buf).digest())
                    buf = b''
                    fs[-1][b'attr'] = b'p'
                else:
                    with open(
                        os.path.join(self.data_path.encode('utf8'), *path), 'rb'
                    ) as _file:
                        r = _file.read(piece_size - len(buf))
                        while r:
                            buf += r
                            if len(buf) == piece_size:
                                pieces.append(sha(buf).digest())
                                # Run the progress function if necessary
                                if progress:
                                    progress(len(pieces), num_pieces)
                                buf = b''
                            else:
                                break
                            r = _file.read(piece_size - len(buf))
            torrent['info']['files'] = fs
            if buf:
                pieces.append(sha(buf).digest())
                if progress:
                    progress(len(pieces), num_pieces)
                buf = ''

        elif os.path.isfile(self.data_path):
            torrent['info']['length'] = get_path_size(self.data_path)
            pieces = []

            with open(self.data_path, 'rb') as _file:
                r = _file.read(piece_size)
                while r:
                    pieces.append(sha(r).digest())
                    if progress:
                        progress(len(pieces), num_pieces)

                    r = _file.read(piece_size)

        torrent['info']['pieces'] = b''.join(pieces)

        # Write out the torrent file
        with open(torrent_path, 'wb') as _file:
            _file.write(bencode(utf8_encode_structure(torrent)))

    def get_data_path(self):
        """Get the path to the files that the torrent will contain.

        Note:
            It can be either a file or a folder.

        Returns:
            str: The torrent data path, either a file or a folder.

        """
        return self.__data_path

    def set_data_path(self, path):
        """Set the path to the files (data) that the torrent will contain.

        Note:
            This property needs to be set before the torrent file can be created and saved.

        Args:
            path (str): The path to the torrent data and can be either a file or a folder.

        Raises:
            InvalidPath: If the path is not found.

        """
        if os.path.exists(path) and (os.path.isdir(path) or os.path.isfile(path)):
            self.__data_path = os.path.abspath(path)
        else:
            raise InvalidPath('No such file or directory: %s' % path)

    def get_piece_size(self):
        """The size of the pieces.

        Returns:
            int: The piece size in multiples of 16 KiBs.
        """
        return self.__piece_size

    def set_piece_size(self, size):
        """Set piece size.

        Note:
            If no piece size is set, one will be automatically selected to
            produce a torrent with less than 1024 pieces or the smallest possible
            with a 8192KiB piece size.

        Args:
            size (int): The desired piece size in multiples of 16 KiBs.

        Raises:
            InvalidPieceSize: If the piece size is not a valid multiple of 16 KiB.

        """
        if size % 16 and size:
            raise InvalidPieceSize('Piece size must be a multiple of 16 KiB')
        self.__piece_size = size

    def get_comment(self):
        """Get the torrent comment.

        Returns:
            str: An informational string about the torrent.

        """
        return self.__comment

    def set_comment(self, comment):
        """Set the comment for the torrent.

        Args:
            comment (str): An informational string about the torrent.

        """
        self.__comment = comment

    def get_private(self):
        """Get the private flag of the torrent.

        Returns:
            bool: True if private flag has been set, else False.

        """
        return self.__private

    def set_private(self, private):
        """Set the torrent private flag.

        Note:
            Private torrents only announce to trackers and will not use DHT or
            Peer Exchange. See http://bittorrent.org/beps/bep_0027.html

        Args:
            private (bool): True if the torrent is to be private.

        """
        self.__private = private

    def get_trackers(self):
        """Get the announce trackers.

        Note:
            See http://bittorrent.org/beps/bep_0012.html

        Returns:
            list of lists: A list containing a list of trackers.

        """
        return self.__trackers

    def set_trackers(self, trackers):
        """Set the announce trackers.

        Args:
            private (list of lists): A list containing lists of trackers as strings, each list is a tier.

        """
        self.__trackers = trackers

    def get_webseeds(self):
        """Get the webseeds.

        Note:
            The web seeds can either be:
                Hoffman-style: http://bittorrent.org/beps/bep_0017.html
                GetRight-style: http://bittorrent.org/beps/bep_0019.html

            If the url ends in '.php' then it will be considered Hoffman-style, if
            not it will be considered GetRight-style.

        Returns:
            list: The webseeds.

        """
        return self.__webseeds

    def set_webseeds(self, webseeds):
        """Set webseeds.

        Note:
            The web seeds can either be:
                Hoffman-style: http://bittorrent.org/beps/bep_0017.html
                GetRight-style: http://bittorrent.org/beps/bep_0019.html

            If the url ends in '.php' then it will be considered Hoffman-style, if
            not it will be considered GetRight-style.

        Args:
            private (list): The webseeds URLs which can be either Hoffman or GetRight style.

        """
        self.__webseeds = webseeds

    def get_pad_files(self):
        """Get status of padding files for the torrent.

        Returns:
            bool: True if padding files have been enabled to align files on piece boundaries.

        """
        return self.__pad_files

    def set_pad_files(self, pad):
        """Enable padding files for the torrent.

        Args:
            private (bool): True adds padding files to align files on piece boundaries.

        """
        self.__pad_files = pad

    data_path = property(get_data_path, set_data_path)
    piece_size = property(get_piece_size, set_piece_size)
    comment = property(get_comment, set_comment)
    private = property(get_private, set_private)
    trackers = property(get_trackers, set_trackers)
    webseeds = property(get_webseeds, set_webseeds)
    pad_files = property(get_pad_files, set_pad_files)



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import os
from base64 import b64encode
from hashlib import sha1 as sha

import pytest
import pytest_twisted
from twisted.internet import defer, reactor, task
from twisted.internet.error import CannotListenError
from twisted.web.http import FORBIDDEN
from twisted.web.resource import EncodingResourceWrapper, Resource
from twisted.web.server import GzipEncoderFactory, Site
from twisted.web.static import File

import deluge.common
import deluge.component as component
import deluge.core.torrent
from deluge._libtorrent import lt
from deluge.conftest import BaseTestCase
from deluge.core.core import Core
from deluge.core.rpcserver import RPCServer
from deluge.error import AddTorrentError, InvalidTorrentError

from . import common

common.disable_new_release_check()


class CookieResource(Resource):
    def render(self, request):
        if request.getCookie(b'password') != b'deluge':
            request.setResponseCode(FORBIDDEN)
            return

        request.setHeader(b'Content-Type', b'application/x-bittorrent')
        with open(
            common.get_test_data_file('ubuntu-9.04-desktop-i386.iso.torrent'), 'rb'
        ) as _file:
            data = _file.read()
        return data


class PartialDownload(Resource):
    def getChild(self, path, request):  # NOQA: N802
        return EncodingResourceWrapper(self, [GzipEncoderFactory()])

    def render(self, request):
        with open(
            common.get_test_data_file('ubuntu-9.04-desktop-i386.iso.torrent'), 'rb'
        ) as _file:
            data = _file.read()
        request.setHeader(b'Content-Length', str(len(data)))
        request.setHeader(b'Content-Type', b'application/x-bittorrent')
        return data


class RedirectResource(Resource):
    def render(self, request):
        request.redirect(b'/ubuntu-9.04-desktop-i386.iso.torrent')
        return b''


class TopLevelResource(Resource):
    def __init__(self):
        Resource.__init__(self)
        self.putChild(b'cookie', CookieResource())
        self.putChild(b'partial', PartialDownload())
        self.putChild(b'redirect', RedirectResource())
        self.putChild(
            b'ubuntu-9.04-desktop-i386.iso.torrent',
            File(common.get_test_data_file('ubuntu-9.04-desktop-i386.iso.torrent')),
        )


class TestCore(BaseTestCase):
    def set_up(self):
        self.rpcserver = RPCServer(listen=False)
        self.core: Core = Core()
        self.core.config.config['lsd'] = False
        self.clock = task.Clock()
        self.core.torrentmanager.clock = self.clock
        self.listen_port = 51242
        return component.start().addCallback(self.start_web_server)

    def start_web_server(self, result):
        website = Site(TopLevelResource())
        for dummy in range(10):
            try:
                self.webserver = reactor.listenTCP(self.listen_port, website)
            except CannotListenError as ex:
                error = ex
                self.listen_port += 1
            else:
                break
        else:
            raise error

        return result

    def tear_down(self):
        def on_shutdown(result):
            del self.rpcserver
            del self.core
            return self.webserver.stopListening()

        return component.shutdown().addCallback(on_shutdown)

    def add_torrent(self, filename, paused=False):
        if not paused:
            # Patch libtorrent flags starting torrents paused
            self.patch(
                deluge.core.torrentmanager,
                'LT_DEFAULT_ADD_TORRENT_FLAGS',
                lt.add_torrent_params_flags_t.flag_auto_managed
                | lt.add_torrent_params_flags_t.flag_update_subscribe
                | lt.add_torrent_params_flags_t.flag_apply_ip_filter,
            )
        options = {'add_paused': paused, 'auto_managed': False}
        filepath = common.get_test_data_file(filename)
        with open(filepath, 'rb') as _file:
            filedump = b64encode(_file.read())
        torrent_id = self.core.add_torrent_file(filename, filedump, options)
        return torrent_id

    @pytest_twisted.inlineCallbacks
    def test_add_torrent_files(self):
        options = {}
        filenames = ['test.torrent', 'test_torrent.file.torrent']
        files_to_add = []
        for f in filenames:
            filename = common.get_test_data_file(f)
            with open(filename, 'rb') as _file:
                filedump = b64encode(_file.read())
            files_to_add.append((filename, filedump, options))
        errors = yield self.core.add_torrent_files(files_to_add)
        assert len(errors) == 0

    @pytest_twisted.inlineCallbacks
    def test_add_torrent_files_error_duplicate(self):
        options = {}
        filenames = ['test.torrent', 'test.torrent']
        files_to_add = []
        for f in filenames:
            filename = common.get_test_data_file(f)
            with open(filename, 'rb') as _file:
                filedump = b64encode(_file.read())
            files_to_add.append((filename, filedump, options))
        errors = yield self.core.add_torrent_files(files_to_add)
        assert len(errors) == 1
        assert str(errors[0]).startswith('Torrent already in session')

    @pytest_twisted.inlineCallbacks
    def test_add_torrent_file(self):
        options = {}
        filename = common.get_test_data_file('test.torrent')
        with open(filename, 'rb') as _file:
            filedump = b64encode(_file.read())
        torrent_id = yield self.core.add_torrent_file_async(filename, filedump, options)

        # Get the info hash from the test.torrent
        from deluge.bencode import bdecode, bencode

        with open(filename, 'rb') as _file:
            info_hash = sha(bencode(bdecode(_file.read())[b'info'])).hexdigest()
        assert torrent_id == info_hash

    def test_add_torrent_file_invalid_filedump(self):
        options = {}
        filename = common.get_test_data_file('test.torrent')
        with pytest.raises(AddTorrentError):
            self.core.add_torrent_file(filename, False, options)

    @pytest_twisted.inlineCallbacks
    def test_add_torrent_url(self, mock_mkstemp):
        url = (
            'http://localhost:%d/ubuntu-9.04-desktop-i386.iso.torrent'
            % self.listen_port
        )
        options = {}
        info_hash = '60d5d82328b4547511fdeac9bf4d0112daa0ce00'

        torrent_id = yield self.core.add_torrent_url(url, options)
        assert torrent_id == info_hash
        assert not os.path.isfile(mock_mkstemp[1])

    @pytest_twisted.ensureDeferred
    async def test_add_torrent_url_with_cookie(self):
        url = 'http://localhost:%d/cookie' % self.listen_port
        options = {}
        headers = {'Cookie': 'password=deluge'}
        info_hash = '60d5d82328b4547511fdeac9bf4d0112daa0ce00'

        with pytest.raises(Exception):
            await self.core.add_torrent_url(url, options)

        result = await self.core.add_torrent_url(url, options, headers)
        assert result == info_hash

    @pytest_twisted.ensureDeferred
    async def test_add_torrent_url_with_redirect(self):
        url = 'http://localhost:%d/redirect' % self.listen_port
        options = {}
        info_hash = '60d5d82328b4547511fdeac9bf4d0112daa0ce00'

        result = await self.core.add_torrent_url(url, options)
        assert result == info_hash

    @pytest_twisted.ensureDeferred
    async def test_add_torrent_url_with_partial_download(self):
        url = 'http://localhost:%d/partial' % self.listen_port
        options = {}
        info_hash = '60d5d82328b4547511fdeac9bf4d0112daa0ce00'

        result = await self.core.add_torrent_url(url, options)
        assert result == info_hash

    @pytest_twisted.inlineCallbacks
    def test_add_torrent_magnet(self):
        info_hash = '60d5d82328b4547511fdeac9bf4d0112daa0ce00'
        tracker = 'udp://tracker.example.com'
        name = 'test magnet'
        uri = deluge.common.create_magnet_uri(info_hash, name=name, trackers=[tracker])
        options = {}
        torrent_id = yield self.core.add_torrent_magnet(uri, options)
        assert torrent_id == info_hash
        torrent_status = self.core.get_torrent_status(torrent_id, ['name', 'trackers'])
        assert torrent_status['trackers'][0]['url'] == tracker
        assert torrent_status['name'] == name

    def test_resume_torrent(self):
        tid1 = self.add_torrent('test.torrent', paused=True)
        tid2 = self.add_torrent('test_torrent.file.torrent', paused=True)
        # Assert paused
        r1 = self.core.get_torrent_status(tid1, ['paused'])
        assert r1['paused']
        r2 = self.core.get_torrent_status(tid2, ['paused'])
        assert r2['paused']

        self.core.resume_torrent(tid2)
        r1 = self.core.get_torrent_status(tid1, ['paused'])
        assert r1['paused']
        r2 = self.core.get_torrent_status(tid2, ['paused'])
        assert not r2['paused']

    def test_resume_torrent_list(self):
        """Backward compatibility for list of torrent_ids."""
        torrent_id = self.add_torrent('test.torrent', paused=True)
        self.core.resume_torrent([torrent_id])
        result = self.core.get_torrent_status(torrent_id, ['paused'])
        assert not result['paused']

    def test_resume_torrents(self):
        tid1 = self.add_torrent('test.torrent', paused=True)
        tid2 = self.add_torrent('test_torrent.file.torrent', paused=True)
        self.core.resume_torrents([tid1, tid2])
        r1 = self.core.get_torrent_status(tid1, ['paused'])
        assert not r1['paused']
        r2 = self.core.get_torrent_status(tid2, ['paused'])
        assert not r2['paused']

    def test_resume_torrents_all(self):
        """With no torrent_ids param, resume all torrents"""
        tid1 = self.add_torrent('test.torrent', paused=True)
        tid2 = self.add_torrent('test_torrent.file.torrent', paused=True)
        self.core.resume_torrents()
        r1 = self.core.get_torrent_status(tid1, ['paused'])
        assert not r1['paused']
        r2 = self.core.get_torrent_status(tid2, ['paused'])
        assert not r2['paused']

    def test_pause_torrent(self):
        tid1 = self.add_torrent('test.torrent')
        tid2 = self.add_torrent('test_torrent.file.torrent')
        # Assert not paused
        r1 = self.core.get_torrent_status(tid1, ['paused'])
        assert not r1['paused']
        r2 = self.core.get_torrent_status(tid2, ['paused'])
        assert not r2['paused']

        self.core.pause_torrent(tid2)
        r1 = self.core.get_torrent_status(tid1, ['paused'])
        assert not r1['paused']
        r2 = self.core.get_torrent_status(tid2, ['paused'])
        assert r2['paused']

    def test_pause_torrent_list(self):
        """Backward compatibility for list of torrent_ids."""
        torrent_id = self.add_torrent('test.torrent')
        result = self.core.get_torrent_status(torrent_id, ['paused'])
        assert not result['paused']
        self.core.pause_torrent([torrent_id])
        result = self.core.get_torrent_status(torrent_id, ['paused'])
        assert result['paused']

    def test_pause_torrents(self):
        tid1 = self.add_torrent('test.torrent')
        tid2 = self.add_torrent('test_torrent.file.torrent')

        self.core.pause_torrents([tid1, tid2])
        r1 = self.core.get_torrent_status(tid1, ['paused'])
        assert r1['paused']
        r2 = self.core.get_torrent_status(tid2, ['paused'])
        assert r2['paused']

    def test_pause_torrents_all(self):
        """With no torrent_ids param, pause all torrents"""
        tid1 = self.add_torrent('test.torrent')
        tid2 = self.add_torrent('test_torrent.file.torrent')

        self.core.pause_torrents()
        r1 = self.core.get_torrent_status(tid1, ['paused'])
        assert r1['paused']
        r2 = self.core.get_torrent_status(tid2, ['paused'])
        assert r2['paused']

    @pytest_twisted.inlineCallbacks
    def test_prefetch_metadata_existing(self):
        """Check another call with same magnet returns existing deferred."""
        magnet = 'magnet:?xt=urn:btih:ab570cdd5a17ea1b61e970bb72047de141bce173'
        expected = ('ab570cdd5a17ea1b61e970bb72047de141bce173', b'')

        d1 = self.core.prefetch_magnet_metadata(magnet)
        d2 = self.core.prefetch_magnet_metadata(magnet)
        dg = defer.gatherResults([d1, d2], consumeErrors=True)
        self.clock.advance(30)
        result = yield dg
        assert result == [expected] * 2

    @pytest_twisted.inlineCallbacks
    def test_remove_torrent(self):
        options = {}
        filename = common.get_test_data_file('test.torrent')
        with open(filename, 'rb') as _file:
            filedump = b64encode(_file.read())
        torrent_id = yield self.core.add_torrent_file_async(filename, filedump, options)
        removed = self.core.remove_torrent(torrent_id, True)
        assert removed
        assert len(self.core.get_session_state()) == 0

    def test_remove_torrent_invalid(self):
        with pytest.raises(InvalidTorrentError):
            self.core.remove_torrent(
                'torrentidthatdoesntexist',
                True,
            )

    @pytest_twisted.inlineCallbacks
    def test_remove_torrents(self):
        options = {}
        filename = common.get_test_data_file('test.torrent')
        with open(filename, 'rb') as _file:
            filedump = b64encode(_file.read())
        torrent_id = yield self.core.add_torrent_file_async(filename, filedump, options)

        filename2 = common.get_test_data_file('unicode_filenames.torrent')
        with open(filename2, 'rb') as _file:
            filedump = b64encode(_file.read())
        torrent_id2 = yield self.core.add_torrent_file_async(
            filename2, filedump, options
        )
        d = self.core.remove_torrents([torrent_id, torrent_id2], True)

        def test_ret(val):
            assert val == []

        d.addCallback(test_ret)

        def test_session_state(val):
            assert len(self.core.get_session_state()) == 0

        d.addCallback(test_session_state)
        yield d

    @pytest_twisted.inlineCallbacks
    def test_remove_torrents_invalid(self):
        options = {}
        filename = common.get_test_data_file('test.torrent')
        with open(filename, 'rb') as _file:
            filedump = b64encode(_file.read())
            torrent_id = yield self.core.add_torrent_file_async(
                filename, filedump, options
            )
        val = yield self.core.remove_torrents(
            ['invalidid1', 'invalidid2', torrent_id], False
        )
        assert len(val) == 2
        assert val[0] == ('invalidid1', 'torrent_id invalidid1 not in session.')
        assert val[1] == ('invalidid2', 'torrent_id invalidid2 not in session.')

    def test_get_session_status(self):
        status = self.core.get_session_status(
            ['net.recv_tracker_bytes', 'net.sent_tracker_bytes']
        )
        assert isinstance(status, dict)
        assert status['net.recv_tracker_bytes'] == 0
        assert status['net.sent_tracker_bytes'] == 0

    def test_get_session_status_all(self):
        status = self.core.get_session_status([])
        assert isinstance(status, dict)
        assert 'upload_rate' in status
        assert 'net.recv_bytes' in status

    def test_get_session_status_depr(self):
        status = self.core.get_session_status(['num_peers', 'num_unchoked'])
        assert isinstance(status, dict)
        assert status['num_peers'] == 0
        assert status['num_unchoked'] == 0

    def test_get_session_status_rates(self):
        status = self.core.get_session_status(['upload_rate', 'download_rate'])
        assert isinstance(status, dict)
        assert status['upload_rate'] == 0

    def test_get_session_status_ratio(self):
        status = self.core.get_session_status(['write_hit_ratio', 'read_hit_ratio'])
        assert isinstance(status, dict)
        assert status['write_hit_ratio'] == 0.0
        assert status['read_hit_ratio'] == 0.0

    def test_get_free_space(self):
        space = self.core.get_free_space('.')
        assert isinstance(space, int)
        assert space >= 0
        assert self.core.get_free_space('/someinvalidpath') == -1

    @pytest.mark.slow
    def test_test_listen_port(self):
        d = self.core.test_listen_port()

        def result(r):
            assert r in (True, False)

        d.addCallback(result)
        return d

    def test_sanitize_filepath(self):
        pathlist = {
            '\\backslash\\path\\': 'backslash/path',
            ' single_file ': 'single_file',
            '..': '',
            '/../..../': '',
            '  Def ////ad./ / . . /b  d /file': 'Def/ad./. ./b  d/file',
            '/ test /\\.. /.file/': 'test/.file',
            'mytorrent/subfold/file1': 'mytorrent/subfold/file1',
            'Torrent/folder/': 'Torrent/folder',
        }

        for key in pathlist:
            assert (
                deluge.core.torrent.sanitize_filepath(key, folder=False)
                == pathlist[key]
            )

            assert (
                deluge.core.torrent.sanitize_filepath(key, folder=True)
                == pathlist[key] + '/'
            )

    def test_get_set_config_values(self):
        assert self.core.get_config_values(['abc', 'foo']) == {'foo': None, 'abc': None}
        assert self.core.get_config_value('foobar') is None
        self.core.set_config({'abc': 'def', 'foo': 10, 'foobar': 'barfoo'})
        assert self.core.get_config_values(['foo', 'abc']) == {'foo': 10, 'abc': 'def'}
        assert self.core.get_config_value('foobar') == 'barfoo'

    def test_read_only_config_keys(self):
        key = 'max_upload_speed'
        self.core.read_only_config_keys = [key]

        old_value = self.core.get_config_value(key)
        self.core.set_config({key: old_value + 10})
        new_value = self.core.get_config_value(key)
        assert old_value == new_value

        self.core.read_only_config_keys = None

    def test__create_peer_id(self):
        assert self.core._create_peer_id('2.0.0') == '-DE200s-'
        assert self.core._create_peer_id('2.0.0.dev15') == '-DE200D-'
        assert self.core._create_peer_id('2.0.1rc1') == '-DE201r-'
        assert self.core._create_peer_id('2.11.0b2') == '-DE2B0b-'
        assert self.core._create_peer_id('2.4.12b2.dev3') == '-DE24CD-'



#
# This file is part of Deluge and is licensed under GNU General Public License 3.0, or later, with
# the additional special exception to link portions of this program with the OpenSSL library.
# See LICENSE for more details.
#

import deluge.component as component
from deluge.conftest import BaseTestCase
from deluge.core.core import Core


class TestAlertManager(BaseTestCase):
    def set_up(self):
        self.core = Core()
        self.core.config.config['lsd'] = False
        self.am = component.get('AlertManager')
        return component.start(['AlertManager'])

    def tear_down(self):
        return component.shutdown()

    def test_register_handler(self):
        def handler(alert):
            return

        self.am.register_handler('dummy_alert', handler)
        assert self.am.handlers['dummy_alert'] == [handler]

    def test_deregister_handler(self):
        def handler(alert):
            return

        self.am.register_handler('dummy_alert', handler)
        self.am.deregister_handler(handler)
        assert self.am.handlers['dummy_alert'] == []
